import logging
import select
import errno
import struct
import argparse
import importlib
import pathlib
import sys
import json
import time
from rich.logging import RichHandler

from pyrtma import Client, Message, MessageHeader, get_msg_cls, ClientError
from pyrtma.exceptions import RTMAMessageError
import pyrtma.core_defs as cd

from socket import error as SocketError
import socket
from socketserver import TCPServer
from websocket_server import (  # type: ignore
    WebsocketServer,
    WebSocketHandler,
    logger,
    OPCODE,
    MASKED,
    FIN,
    PAYLOAD_LEN,
    OPCODE_CLOSE_CONN,
    OPCODE_BINARY,
    OPCODE_CONTINUATION,
    OPCODE_PING,
    OPCODE_PONG,
    OPCODE_TEXT,
)

from typing import Optional, List, Callable, Any, Dict, cast

# use rich logging
logging.getLogger().removeHandler(logging.getLogger().handlers[0])
logging.getLogger().addHandler(RichHandler())


class RTMAWebSocketHandler(WebSocketHandler):
    """RTMA Web Socket Handler class"""

    disable_nagle_algorithm = True

    def __init__(self, socket, addr, server):
        """RTMA Web Socket Handler

        Initializes and handles RTMA Proxy connection

        Args:
            socket: socket object
            addr: client address
            server: server object
        """
        # Initialize RTMA Proxy connection
        self.mm_ip = server.mm_ip
        self.proxy = Client()

        WebSocketHandler.__init__(self, socket, addr, server)

    @property
    def ws_ready_to_send(self):
        if self.keep_alive:
            _, wd, _ = select.select([], [self.wfile], [], 0)
            return self.wfile in wd
        else:
            return False

    def handle(self):
        """Handle RTMA proxy connection"""
        if not self.handshake_done:
            self.handshake()

        if not self.valid_client:
            logger.error("Websocket handshake failed.")
            return

        # Establish the underlying socket connection with MessageManger Server
        self.proxy._socket_connect(self.mm_ip)

        log_str = (
            "New Client:\n"
            + f"ws: {self.request.getsockname()[0]}:{self.request.getsockname()[1]} -> {self.request.getpeername()[0]}:{self.request.getpeername()[1]}\n"
            + f"rtma: {self.proxy.sock.getsockname()[0]}:{self.proxy.sock.getsockname()[1]} -> {self.proxy.sock.getpeername()[0]}:{self.proxy.sock.getpeername()[1]}\n"
        )
        logger.info(log_str)

        # Message Loop
        while self.keep_alive and self.proxy.connected:
            rd, _, _ = select.select([self.rfile, self.proxy.sock], [], [], 0.100)

            if self.rfile in rd:
                self.read_ws_message()

            if self.proxy.sock in rd and self.proxy.connected:
                try:
                    msg = self.proxy.read_message(timeout=None, ack=True)
                except RTMAMessageError as e:
                    error_json = json.dumps(
                        {"rtma_msg_error": str(e)}, separators=(",", ":")
                    )
                    if self.ws_ready_to_send:
                        self.send_message(error_json)
                    logger.error(e, stack_info=False)
                    continue

                if msg is not None and self.keep_alive:
                    # Pass message thru websocket as json
                    if self.ws_ready_to_send:
                        logger.debug(
                            f"Forwarding message type {get_msg_cls(msg.header.msg_type).type_name} to ws"
                        )
                        self.send_message(msg.to_json(minify=True))
                    else:
                        if self.proxy.connected:
                            self.send_failed_message(msg.header, time.perf_counter())
                        logger.warning(
                            f"Failed to foward message type {get_msg_cls(msg.header.msg_type).type_name} to ws. Mod ID = {self.proxy.module_id}"
                        )

    def pong_received(self, msg: str):
        """Log that pong was received

        Args:
            msg (str): Ignored
        """
        logger.info("Websocket PONG received.")

    def handle_connect(self, msg: Message):
        logger.debug("Received CONNECT")

        # self.proxy._module_id = msg.header.src_mod_id
        # self.proxy._host_id = msg.header.src_host_id
        msg.data = cast(cd.MDF_CONNECT, msg.data)
        ack_msg = self.proxy._connect_helper(
            bool(msg.data.logger_status),
            bool(msg.data.daemon_status),
            allow_multiple=False,
        )
        # forward ack
        # Pass message thru websocket as json
        if self.ws_ready_to_send:
            logger.debug(
                f"Forwarding ACK from connect to ws. Mod ID = {self.proxy.module_id}"
            )
            self.send_message(ack_msg.to_json(minify=True))
        else:
            self.send_failed_message(ack_msg.header, time.perf_counter())
            logger.warning(
                f"Failed to foward ACK to ws. Mod ID = {self.proxy.module_id}."
            )

    def handle_connect_v2(self, msg: Message):
        logger.debug("Received CONNECT_V2")

        # self.proxy._module_id = msg.header.src_mod_id
        # self.proxy._host_id = msg.header.src_host_id
        msg.data = cast(cd.MDF_CONNECT_V2, msg.data)
        self.proxy._name = msg.data.name
        ack_msg = self.proxy._connect_helper(
            bool(msg.data.logger_status),
            bool(msg.data.daemon_status),
            bool(msg.data.allow_multiple),  # TODO add CONNECT_V2 support to rtma-js
        )
        # forward ack
        # Pass message thru websocket as json
        if self.ws_ready_to_send:
            logger.debug(
                f"Forwarding ACK from connect to ws. Mod ID = {self.proxy.module_id}"
            )
            self.send_message(ack_msg.to_json(minify=True))
        else:
            self.send_failed_message(ack_msg.header, time.perf_counter())
            logger.warning(
                f"Failed to foward ACK to ws. Mod ID = {self.proxy.module_id}."
            )

    def process_json_message(self, message: str):
        """Process incoming json message

        Called when a client receives a message over websocket

        Args:
            message (str): JSON message string
        """
        if message == "PING" and self.ws_ready_to_send:
            self.send_message("PONG")
            return

        try:
            msg = Message.from_json(message)

            if msg.header.msg_type == cd.MT_DISCONNECT:
                self.proxy.disconnect()
                logger.info("Received disconnect, disconnected proxy from MM.")
            elif msg.header.msg_type == cd.MT_SUBSCRIBE:
                msg.data = cast(cd.MDF_SUBSCRIBE, msg.data)
                self.proxy.subscribe([msg.data.msg_type])
                logger.info(
                    f"Received SUBSCRIBE: {get_msg_cls(msg.data.msg_type).type_name}"
                )
            elif msg.header.msg_type == cd.MT_UNSUBSCRIBE:
                msg.data = cast(cd.MDF_UNSUBSCRIBE, msg.data)
                self.proxy.unsubscribe([msg.data.msg_type])
                logger.info(
                    f"Received UNSUBSCRIBE: {get_msg_cls(msg.data.msg_type).type_name}"
                )
            elif msg.header.msg_type == cd.MT_PAUSE_SUBSCRIPTION:
                msg.data = cast(cd.MDF_PAUSE_SUBSCRIPTION, msg.data)
                self.proxy.pause_subscription([msg.data.msg_type])
                logger.info(
                    f"Received PAUSE_SUBSCRIPTION: {get_msg_cls(msg.data.msg_type).type_name}"
                )
            elif msg.header.msg_type == cd.MT_RESUME_SUBSCRIPTION:
                msg.data = cast(cd.MDF_RESUME_SUBSCRIPTION, msg.data)
                self.proxy.resume_subscription([msg.data.msg_type])
                logger.info(
                    f"Received RESUME_SUBSCRIPTION: {get_msg_cls(msg.data.msg_type).type_name}"
                )
            elif msg.header.msg_type == cd.MT_CONNECT:
                self.handle_connect(msg)
                logger.info(
                    f"Proxy CONNECTED, assigned module ID {self.proxy.module_id}"
                )
            elif msg.header.msg_type == cd.MT_CONNECT_V2:
                self.handle_connect_v2(msg)
                logger.info(
                    f"Proxy CONNECTED, assigned module ID {self.proxy.module_id}"
                )
            else:
                self.proxy.forward_message(msg.header, msg.data or None)
                logger.debug(
                    f"Forwarded message type {get_msg_cls(msg.header.msg_type).type_name} from ws"
                )
        except (RTMAMessageError, ClientError) as e:
            if self.ws_ready_to_send:
                error_json = json.dumps(
                    {"rtma_msg_error": str(e)}, separators=(",", ":")
                )
                self.send_message(error_json)
            logger.error(e, stack_info=False)
            return

    def read_ws_message(self) -> Optional[str]:
        """Read websocket message

        Returns:
            Optional[str]: Websocket message string
        """
        try:
            b1, b2 = self.read_bytes(2)
        except SocketError as e:  # to be replaced with ConnectionResetError for py3
            if e.errno == errno.ECONNRESET:
                logger.info("Client closed connection.")
                self.keep_alive = 0
                return None
            b1, b2 = 0, 0
        except ValueError as e:
            b1, b2 = 0, 0

        fin = b1 & FIN
        opcode = b1 & OPCODE
        masked = b2 & MASKED
        payload_length = b2 & PAYLOAD_LEN
        opcode_handler: Callable[[str], None]

        if opcode == OPCODE_CLOSE_CONN:
            logger.info("Client asked to close connection.")
            self.keep_alive = 0
            return None
        if not masked:
            logger.warning("Client must always be masked.")
            self.keep_alive = 0
            return None
        if opcode == OPCODE_CONTINUATION:
            logger.warning("Continuation frames are not supported.")
            return None
        elif opcode == OPCODE_BINARY:
            logger.warning("Binary frames are not supported.")
            return None
        elif opcode == OPCODE_TEXT:
            opcode_handler = self.process_json_message
        elif opcode == OPCODE_PING:
            opcode_handler = self.send_pong
        elif opcode == OPCODE_PONG:
            opcode_handler = self.pong_received
        else:
            logger.warning("Unknown opcode %#x." % opcode)
            self.keep_alive = 0
            return None

        if payload_length == 126:
            payload_length = struct.unpack(">H", self.rfile.read(2))[0]
        elif payload_length == 127:
            payload_length = struct.unpack(">Q", self.rfile.read(8))[0]

        masks = self.read_bytes(4)
        message_bytes = bytearray()
        for message_byte in self.read_bytes(payload_length):
            message_byte ^= masks[len(message_bytes) % 4]
            message_bytes.append(message_byte)

        opcode_handler(message_bytes.decode("utf8"))
        return message_bytes.decode("utf8")

    def send_failed_message(self, header: MessageHeader, time_of_failure: float):
        """Send FAILED_MESSAGE when we cannot forward to websocket"""
        data = cd.MDF_FAILED_MESSAGE()
        data.dest_mod_id = self.proxy.module_id
        data.time_of_failure = time_of_failure

        # Copy the values into the RTMA_MSG_HEADER
        for fname, ftype, *_ in data.msg_header._fields_:
            setattr(data.msg_header, fname, getattr(header, fname))

        if (
            data.msg_header.msg_type == cd.MT_FAILED_MESSAGE
        ):  # avoid unlikely infinite recursion
            return
        if self.proxy.connected:
            self.proxy.send_message(data)

    def finish(self):
        """Close RTMA connection"""
        # Disconnect on behalf of the web client if still connected here
        if self.proxy.connected:
            self.proxy.disconnect()
            logger.debug("Disconnected proxy from MM.")

        super().finish()


class WebMessageManager(WebsocketServer):
    """WebMessageManager class"""

    def __init__(
        self,
        host="",
        port=0,
        mm_ip: str = "127.0.0.1:7111",
        loglevel: int = logging.WARNING,
        key=None,
        cert=None,
    ):
        """WebMessageManager class

        Args:
            host (str, optional): IP for WebMessageManager to listen for connections. Defaults to "" (any local IP).
            port (int, optional): Port for WebMessageManager to bind to. Defaults to 0.
            mm_ip (str, optional): Address for RTMA MessageManager. Defaults to "127.0.0.1:7111".
            loglevel (int, optional): Loging level. Defaults to logging.WARNING.
            key (optional): Path to SSL key. Defaults to None.
            cert (optional): Path to SSL cert. Defaults to None.
        """

        logger.setLevel(loglevel)
        TCPServer.__init__(self, (host, port), RTMAWebSocketHandler)
        self.host = host
        self.port = self.socket.getsockname()[1]

        if self.host not in ("", "0.0.0.0"):
            logger.info(f"WebMessageManager listening on {self.host}:{self.port}")
        else:
            hosts = socket.gethostbyname_ex(socket.gethostname())[-1]
            if "127.0.0.1" not in hosts:
                hosts.append("127.0.0.1")
            logger.info(
                f"WebMessageManager listening on all interfaces, port {self.port}"
            )
            logger.info(f"IP addresses: {hosts}")

        self.key = key
        self.cert = cert

        self.clients: List[dict] = []
        self.id_counter = 0
        self.thread = None

        self._deny_clients = False

        self.set_fn_new_client(ws_client_connect)
        self.set_fn_client_left(ws_client_disconnect)

        self.mm_ip = mm_ip


def ws_client_connect(client: Dict[str, Any], server: WebMessageManager):
    """Websocket client connect

        Called for every client connecting (after handshake)
    Args:
        client (Dict[str, Any]): Client dictionary
        server: WebMessageManager Server object
    """
    try:
        logger.info(f"Client connected -> id: {client['id']}")
    except:
        pass


def ws_client_disconnect(client: Dict[str, Any], server: WebMessageManager):
    """Websocket client disconnect

        Called for every client disconnecting
    Args:
        client (Dict[str, Any]): Client dictionary
        server: WebMessageManager Server object
    """
    try:
        logger.info(f"Client disconnected -> id: {client['id']}")
    except:
        pass


def main():
    """Main function for starting web_manager"""

    parser = argparse.ArgumentParser(description="Websocket Message Manager")

    parser.add_argument(
        "-m",
        "--mm-ip",
        default="127.0.0.1:7111",
        dest="mm_ip",
        type=str,
        help="IP address of Message Manager. Defaults to 127.0.0.1:7111.",
    )

    parser.add_argument(
        "-a",
        "--host",
        default="",
        dest="host",
        type=str,
        help='Host IP address, defaults to "" (any local IP).',
    )

    parser.add_argument(
        "-p",
        "--port",
        default=5678,
        dest="port",
        type=int,
        help="Port to listen for websocket clients. Defaults to 5678.",
    )

    parser.add_argument(
        "-d",
        "--defs",
        dest="defs_file",
        type=str,
        help="Path to python message definitions file. Required argument.",
    )

    if len(sys.argv) == 1:
        parser.print_help(sys.stderr)
        sys.exit(1)
    args = parser.parse_args()

    websocket_server = WebMessageManager(
        host=args.host,
        port=args.port,
        loglevel=logging.INFO,
        mm_ip=args.mm_ip,
    )

    base = pathlib.Path(args.defs_file).absolute().parent
    fname = pathlib.Path(args.defs_file).stem

    sys.path.insert(0, (str(base.absolute())))
    importlib.import_module(fname)

    websocket_server.run_forever()


if __name__ == "__main__":
    main()



from typing import ClassVar
from .message_base import MessageBase, MessageMeta


class MessageData(MessageBase, metaclass=MessageMeta):
    """MessageData base class

    This is intended to be treated as an abstract class and
    and should not be directly instantiated.
    """

    type_id: ClassVar[int] = -1
    type_name: ClassVar[str] = ""
    type_hash: ClassVar[int]
    type_source: ClassVar[str] = ""
    type_size: ClassVar[int] = -1
    type_def: ClassVar[str] = ""



import pathlib
import ctypes
import os
import sys
import importlib
import warnings
import pyrtma
import pyrtma.message
import pyrtma.context

from typing import List, Union, Generator, Dict, Any, Optional, Type

from ..context import RTMAContext
from ..validators import ByteArray, Uint32, String
from ..message import Message, MessageHeader, MessageData
from ..message_base import MessageBase, MessageMeta
from ..exceptions import VersionMismatchWarning


_unknown: Dict[int, Type[MessageData]] = {}


def legacy_variable_len_msg(msg_type: int) -> bool:
    var_msgs = set([31, 33, 34, 35, 37, 38, 46, 47, 64, 66, 67, 68, 77, 91, 93])
    return msg_type in var_msgs


def create_unknown(hdr: MessageHeader, raw: bytes) -> MessageData:
    if legacy_variable_len_msg(hdr.msg_type):
        name = f"LegacyMessageDef_{hdr.msg_type:04d}"
        cls = _unknown.get(hdr.msg_type)
        if cls is None:
            cls = type(
                name,
                (MessageData,),
                dict(
                    metaclass=MessageMeta,
                    type_name=name,
                    type_id=hdr.msg_type,
                    type_size=max(hdr.num_data_bytes, len(raw)),
                    var_length_text=String(256),
                ),
            )

        obj = cls()
        obj.var_length_text = raw[: min(256, len(raw))].decode()

    else:
        if len(raw) == hdr.num_data_bytes:
            name = f"UnknownMessageDef_{hdr.msg_type:04d}"
        else:
            name = f"InvalidMessageDef_{hdr.msg_type:04d}"

        cls = _unknown.get(hdr.msg_type)
        if cls is None:
            cls = type(
                name,
                (MessageData,),
                dict(
                    metaclass=MessageMeta,
                    type_name=name,
                    type_id=hdr.msg_type,
                    type_size=max(hdr.num_data_bytes, len(raw)),
                    raw=ByteArray(max(hdr.num_data_bytes, len(raw))),
                ),
            )

        obj = cls()
        obj.raw[:] = raw

    _unknown[hdr.msg_type] = cls

    return obj


class QLFileHeader(MessageBase, metaclass=MessageMeta):
    format_version: Uint32 = Uint32()
    total_bytes: Uint32 = Uint32()
    num_messages: Uint32 = Uint32()
    message_header_size: Uint32 = Uint32()
    data_block_offset_size: Uint32 = Uint32()
    num_data_bytes: Uint32 = Uint32()


class QLReader:
    def __init__(self) -> None:
        self.file_path: Optional[pathlib.Path] = None
        self.defs_path: Optional[pathlib.Path] = None
        self.file_header = QLFileHeader()
        self.headers: List[MessageHeader] = []
        self.offsets: List[int] = []
        self.data: List[MessageData] = []
        self.messages: List[Message] = []
        self.context: RTMAContext = RTMAContext()
        self.skipped = 0

    def clear(self):
        self.file_path = None
        self.defs_path = None
        self.file_header = QLFileHeader()
        self.headers.clear()
        self.offsets.clear()
        self.data.clear()
        self.context = RTMAContext()
        self.messages.clear()
        self.skipped = 0

    def load(
        self,
        binfile: Union[str, os.PathLike],
        msgdefs: Union[str, os.PathLike],
        skip_unknown: bool = True,
    ):
        self.clear()
        self.defs_path = pathlib.Path(msgdefs)
        self.file_path = pathlib.Path(binfile)

        # Import the message definitions
        base = self.defs_path.absolute().parent
        fname = self.defs_path.stem

        # Copy the current message def context before importing
        defs = pyrtma.message._get_msg_defs()
        ctx = pyrtma.context.get_context()

        # Load global message defs
        mt_to_mdf_global = {v.type_id: v for v in self.context.MDF.values()}

        sys.path.insert(0, (str(base.absolute())))
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", VersionMismatchWarning)
            if fname in sys.modules:
                mod = importlib.reload(sys.modules[fname])
            else:
                mod = importlib.import_module(fname)

        # Load local message defs
        ctx2 = mod.get_context()  ####_create_context() (older version)
        if isinstance(ctx2, dict):
            mt_to_mdf_session = {
                ctx2["mt"][mtn]: ctx2["mdf"][mtn] for mtn in ctx2["mdf"]
            }
        elif hasattr(ctx2, "MT"):
            mt_to_mdf_session = {ctx2.MT[mtn]: ctx2.MDF[mtn] for mtn in ctx2.MDF.keys()}
        else:
            print("Warning no local message defs loaded.")

        try:
            messages = []
            headers = []
            data = []

            with open(self.file_path, "rb") as f:
                # Parse binary file header
                file_header = QLFileHeader.from_buffer_copy(
                    f.read(ctypes.sizeof(QLFileHeader))
                )
                msg_header_size = file_header.message_header_size

                # Extract the message headers
                for _ in range(file_header.num_messages):
                    raw = f.read(msg_header_size)
                    headers.append(MessageHeader.from_buffer_copy(raw))

                # Extract the message data offsets for each message
                offset_size = file_header.data_block_offset_size
                offsets = (ctypes.c_uint32 * file_header.num_messages).from_buffer_copy(
                    f.read(offset_size * file_header.num_messages)
                )
                self.offsets = list(map(int, offsets))

                # Read the entire data block remaining
                d_bytes = f.read()

                unknown: List[int] = []
                # Extract the message data for each message
                for n, offset in enumerate(offsets):
                    header = headers[n]

                    # Look for message def by searching mt in local first, then global
                    msg_cls = None
                    if header.msg_type in mt_to_mdf_session:
                        msg_cls = mt_to_mdf_session[header.msg_type]
                    elif header.msg_type in mt_to_mdf_global:
                        msg_cls = mt_to_mdf_global[header.msg_type]

                    raw_bytes = d_bytes[offset : offset + header.num_data_bytes]

                    if msg_cls is None:
                        # print(f"Unknown message definition: MT={header.msg_type}")
                        msg_data = create_unknown(header, raw_bytes)
                        unknown.append(n)

                    elif msg_cls.type_size != header.num_data_bytes:
                        print(
                            f"Warning: Message header indicates a message data size ({header.num_data_bytes}) that does not match the expected size of message type {msg_cls.type_name} "
                            f"({msg_cls.type_size}). Message definitions may be out of sync."
                        )
                        msg_data = create_unknown(header, raw_bytes)
                        unknown.append(n)
                    else:
                        msg_data = msg_cls.from_buffer_copy(raw_bytes)

                    data.append(msg_data)
                    messages.append(Message(header, msg_data))

        finally:
            # Restore the orignal message def context
            pyrtma.message._set_msg_defs(defs)
            pyrtma.context._set_context(ctx)

        # Store the results in the object
        self.file_header = file_header
        self.headers = []
        self.data = []
        self.messages = []
        if skip_unknown:
            self.skipped = len(unknown)
            for n, (h, d, m) in enumerate(zip(headers, data, messages, strict=True)):
                if n not in unknown:
                    self.headers.append(h)
                    self.data.append(d)
                    self.messages.append(m)

        else:
            self.headers.extend(headers)
            self.data.extend(data)
            self.messages.extend(messages)

    def export_json(self, file: Union[str, os.PathLike]):
        save_path = pathlib.Path(file)

        if save_path.exists():
            raise FileExistsError(f"File already exists at: {save_path}")

        if self.file_header.num_messages == 0:
            raise RuntimeError("QLReader does not contain any messages")

        # Write each message json object on a separate line
        with open(save_path, "w") as f:
            for msg in self.messages:
                # Open json object
                f.write("{")

                # Header json object
                f.write(msg.to_json(minify=True))

                # Close json object
                f.write("}\n")

    def iter_by_mt(self, msg_type: int) -> Generator[Message, Any, None]:
        for msg in self.messages:
            if msg.type_id == msg_type:
                yield (msg)

    def iter_by_src(
        self,
        mod_id: int,
    ) -> Generator[Message, Any, None]:
        for msg in self.messages:
            if msg.header.src_mod_id == mod_id:
                yield msg



import pyrtma

from ..data_formatter import DataFormatter
from typing import ClassVar


class RawFormatter(DataFormatter):
    name: ClassVar[str] = "raw"
    mode: ClassVar[str] = "wb"
    ext: ClassVar[str] = ".raw"

    def format_message(self, msg: pyrtma.Message) -> bytes:
        return bytes(msg.header) + bytes(msg.data)



"""pyrtma.compile Message Type Compiler"""

import pathlib
import re
import sys
from typing import List, Union, Dict

from .parser import Parser, ParserError, FileFormatError
from rich.traceback import install
import warnings
from pyrtma.compilers.python import PyDefCompiler
from pyrtma.compilers.javascript import JSDefCompiler
from pyrtma.compilers.matlab import MatlabDefCompiler
from pyrtma.compilers.c99 import CDefCompiler
from pyrtma.compilers.yaml import YAMLCompiler
from pyrtma.compilers.info import InfoCompiler

install(word_wrap=True, show_locals=True)


def compile(
    defs_files: List[str],  # str, # change back to str after removing v1 compiler
    out_dir: str,
    out_name: str,
    python: bool = False,
    javascript: bool = False,
    matlab: bool = False,
    c_lang: bool = False,
    info: bool = False,
    combined: bool = False,
    debug: bool = False,
    validate_alignment: bool = True,
    auto_pad: bool = True,
    import_coredefs: bool = True,
):
    """compile message defs

    Args:
        defs_files (List[str]): Root YAML message defintion file to parse. List of C header file(s) will use v1 python compiler (deprecated)
        out_name (str): Output directory for compiled files. For v1 compiler (deprecated), full output filename.
        python (bool, optional): Output python .py file. Defaults to False.
        javascript (bool, optional): Output javascript .js file. Defaults to False.
        matlab (bool, optional): Output matlab .m file. Defaults to False.
        c_lang (bool, optional): Output C .h file. Defaults to False.
        info (bool, optional): Output info .txt file. Defaults to False.
        combined (bool, optional): Output combined YAML file. Defaults to False.
        debug (bool, optional): Debug mode. Defaults to False.
        validate_alignment(bool, optional): Validate message 64-bit alignment. Defaults to True.
        auto_pad(bool, optional): Automatically pad messages failing 64-bit alignment validation. Defaults to True. Has no effect if validate_alignment is False.
        import_coredefs(bool, optional): Automatically import pyrtma.core_defs. Defaults to True.


    Raises:
        FileFormatError: Issue with input file format
        FileExistsError: Output file is not a directory
        RuntimeError: Invalid output filename
    """
    # determine if using v1 or v2 compiler
    file1_ext = pathlib.Path(defs_files[0]).suffix
    if file1_ext.lower() == ".h":
        compiler_version = 1
        warnings.warn(
            "V1 message def .h compiler is deprecated and has been replaced by the V2 yaml compiler.",
            FutureWarning,
        )
    elif file1_ext.lower() in [".yaml", ".yml"]:
        compiler_version = 2
        if len(defs_files) > 1:
            raise FileFormatError("defs_file must be a single .yaml file")
    else:
        raise FileFormatError("Unexpected file type. defs_file must be a .yaml file.")

    if compiler_version == 1:
        from pyrtma.compilers.python_v1 import python_v1_compile

        print("Building V1 python message definitions...")
        python_v1_compile(include_files=defs_files, out_filename=out_dir)
        print("DONE.")
        return
    # else continue with compiler V2

    defs_file = defs_files[0]
    parser = Parser(
        debug=debug,
        validate_alignment=validate_alignment,
        auto_pad=auto_pad,
        import_coredefs=import_coredefs,
    )
    parser.parse(pathlib.Path(defs_file))

    if debug:
        with open("parser.json", "w") as f:
            f.write(parser.to_json())

    # Use the same directory and filename as the message defs file by default
    if out_dir == "":
        outpath = pathlib.Path(defs_file).parent.resolve().absolute()
    else:
        outpath = pathlib.Path(out_dir)

    if not outpath.is_dir():
        raise FileExistsError(f"The output location must be a directory.")

    # Make the final output directory, but no anything above it.
    outpath.mkdir(parents=False, exist_ok=True)

    if out_name == "":
        filename = pathlib.Path(defs_file).stem
    else:
        filename = out_name

    # Check for any invalid names
    if re.search(r"[^a-zA-Z0-9_-]", filename):
        raise RuntimeError(f"Invalid out filename: {filename}")

    compiler: Union[
        PyDefCompiler,
        JSDefCompiler,
        MatlabDefCompiler,
        CDefCompiler,
        YAMLCompiler,
        InfoCompiler,
    ]
    if python:
        print("Building python message definitions...")

        compiler = PyDefCompiler(parser, debug=debug)
        ext = ".py"
        output = outpath / (filename + ext)
        compiler.generate(output)

    if javascript:
        print("Building javascript message definitions...")

        compiler = JSDefCompiler(parser, debug=debug)
        ext = ".js"
        output = outpath / (filename + ext)
        compiler.generate(output)

    if matlab:
        print("Building matlab message definitions...")

        compiler = MatlabDefCompiler(parser, debug=debug)
        ext = ".m"
        output = outpath / (filename + ext)
        compiler.generate(output)

    if c_lang:
        print("Building C/C++ message definitions...")

        compiler = CDefCompiler(parser, filename=filename, debug=debug)
        ext = ".h"
        output = outpath / (filename + ext)
        compiler.generate(output)

    if combined:
        print("Building combined yaml file...")

        compiler = YAMLCompiler(parser, filename=filename, debug=debug)
        ext = ".yaml"
        output = outpath / (filename + "_combined" + ext)
        compiler.generate(output)

    if info:
        print("Building info file...")

        compiler = InfoCompiler(parser, filename=filename, debug=debug)
        ext = ".txt"
        output = outpath / (filename + ext)
        compiler.generate(output)

    print("DONE.")


def main() -> None:
    import argparse

    # subparser for include file, to parse compiler options first
    argparser1 = argparse.ArgumentParser(
        description="pyrtma Message Definition Compiler.",
        add_help=False,
    )

    argparser1.add_argument(
        "-i",
        "-I",
        "--defs",
        nargs="*",  # for backwards compatibility with v1 compiler. Remove in future version along with v1 compiler.
        dest="defs_files",
        help="YAML message defintion file to parse. C header file(s) will use v1 python compiler (deprecated)",
    )

    # full arg parser
    argparser = argparse.ArgumentParser(
        parents=[argparser1],
        conflict_handler="resolve",
        add_help=True,
    )

    argparser.add_argument(
        "--c",
        "--c_lang",
        dest="c_lang",
        action="store_true",
        help="Output C .h file",
    )

    argparser.add_argument(
        "--python",
        "--py",
        dest="python",
        action="store_true",
        help="Output python .py file",
    )

    argparser.add_argument(
        "--javascript",
        "--js",
        dest="javascript",
        action="store_true",
        help="Output javascript .js file",
    )

    argparser.add_argument(
        "--matlab",
        "--mat",
        "--ml",
        dest="matlab",
        action="store_true",
        help="Output matlab .m file",
    )

    argparser.add_argument(
        "--info",
        dest="info",
        action="store_true",
        help="Output info .txt file",
    )

    argparser.add_argument(
        "--combined",
        dest="combined",
        action="store_true",
        help="Output combined yaml file",
    )

    argparser.add_argument(
        "-o",
        "--out",
        default="",
        dest="out_dir",
        help="Output directory for compiled files. For v1 compiler (deprecated), full output filename.",
    )

    argparser.add_argument(
        "-n",
        "--name",
        default="",
        dest="out_name",
        help="Output file(s) base name.",
    )

    argparser.add_argument(
        "--debug",
        dest="debug",
        action="store_true",
        help="Debug compiler",
    )

    # parse input file first to parse compiler opts in yaml file
    args_temp, _ = argparser1.parse_known_args()

    # Default compiler options
    COMPILER_OPTS: Dict[str, Union[int, float, bool, str]] = {
        "IMPORT_COREDEFS": True,
        "VALIDATE_ALIGNMENT": True,
        "AUTO_PAD": True,
    }
    # parse yaml compiler_options and replace defaults prior to parsing command line args
    try:
        defs_files = args_temp.defs_files
        if (
            defs_files
            and len(defs_files) == 1
            and pathlib.Path(defs_files[0]).suffix.lower()
            in [
                ".yaml",
                ".yml",
            ]
        ):
            defs_file = defs_files[0]
            parser = Parser()
            compiler_options = parser.parse_compiler_options(pathlib.Path(defs_file))
            for key, value in compiler_options.items():
                COMPILER_OPTS[key] = value.value
    except (ParserError, FileNotFoundError) as e:
        print()
        msg = " ".join(str(arg) for arg in e.args)
        print(f"{e.__class__.__name__}: {msg}")
        print()
        if e.__cause__:
            print("Details:")
            msg = " ".join(str(arg) for arg in e.__cause__.args)
            print(f"\t{e.__cause__.__class__.__name__}: {msg}")
        sys.exit(1)
    except Exception:
        raise

    # defualt compiler opts to values from yaml
    argparser.add_argument(
        "--no_val_align",
        dest="validate_alignment",
        action="store_false",
        default=COMPILER_OPTS["VALIDATE_ALIGNMENT"],
        help="Disable 64-bit alignment validation",
    )

    argparser.add_argument(
        "--no_auto_pad",
        dest="auto_pad",
        action="store_false",
        default=COMPILER_OPTS["AUTO_PAD"],
        help="Disable 64-bit alignment auto-padding",
    )

    argparser.add_argument(
        "--no_core_import",
        dest="import_coredefs",
        action="store_false",
        default=COMPILER_OPTS["IMPORT_COREDEFS"],
        help="Disable auto-import of pyrtma.core_defs",
    )

    # parse all options
    args = argparser.parse_args()

    try:
        # run compiler
        compile(**vars(args))
    except (ParserError, FileNotFoundError) as e:
        print()
        msg = " ".join(str(arg) for arg in e.args)
        print(f"{e.__class__.__name__}: {msg}")
        print()
        if e.__cause__:
            print("Details:")
            msg = " ".join(str(arg) for arg in e.__cause__.args)
            print(f"\t{e.__cause__.__class__.__name__}: {msg}")
        sys.exit(1)
    except Exception:
        raise

    sys.exit(0)


if __name__ == "__main__":
    main()



import logging
import logging.handlers
import pyrtma.core_defs as cd
import pathlib
import weakref

from .message import MessageData
from abc import ABC, abstractmethod
from .exceptions import ClientError

from typing import Union, Type, Dict, Optional, List
from contextlib import contextmanager
import traceback
from rich.logging import RichHandler
from rich.markup import escape


class ClientLike(ABC):
    """Abstract base class for ClientLike classes such as Client and MessageManager

    Alternative to typing.Protocol for older python versions"""

    @property
    @abstractmethod
    def connected(self) -> bool: ...

    @property
    @abstractmethod
    def logger(self) -> "RTMALogger": ...

    @abstractmethod
    def send_message(
        self,
        msg_data: MessageData,
        dest_mod_id: int = 0,
        dest_host_id: int = 0,
        timeout: float = -1,
    ) -> None: ...


# TODO after dropping support for py3.7, switch to Protocol and remove abstract base class from Client and MessageManager
# class ClientLike(Protocol):
#     def send_message(
#         self,
#         msg_data: MessageData,
#         dest_mod_id: int = 0,
#         dest_host_id: int = 0,
#         timeout: float = -1,
#     ) -> None: ...

#     @property
#     def connected(self) -> bool: ...

#     @property
#     def logger(self) -> "RTMALogger": ...


RTMA_LOG_MSG = Union[
    cd.MDF_RTMA_LOG,
    cd.MDF_RTMA_LOG_DEBUG,
    cd.MDF_RTMA_LOG_INFO,
    cd.MDF_RTMA_LOG_WARNING,
    cd.MDF_RTMA_LOG_ERROR,
    cd.MDF_RTMA_LOG_CRITICAL,
]


class RTMALogHandler(logging.Handler):
    """Logging handler class that writes logs as rtma messages"""

    log_map: Dict[int, Type[RTMA_LOG_MSG]] = {
        10: cd.MDF_RTMA_LOG_DEBUG,
        20: cd.MDF_RTMA_LOG_INFO,
        30: cd.MDF_RTMA_LOG_WARNING,
        40: cd.MDF_RTMA_LOG_ERROR,
        50: cd.MDF_RTMA_LOG_CRITICAL,
    }

    def __init__(self, client_ref: "weakref.ReferenceType[ClientLike]"):
        logging.Handler.__init__(self)
        self.client_ref = client_ref

    def close(self):
        logging.Handler.close(self)

    def get_log_msg_cls(self, level) -> Type[RTMA_LOG_MSG]:
        return RTMALogHandler.log_map.get(level) or cd.MDF_RTMA_LOG

    def gen_log_msg(self, record: logging.LogRecord) -> RTMA_LOG_MSG:
        # See https://docs.python.org/3/library/logging.html#logrecord-attributes
        # for LogRecord attributes that could be added
        msg = self.get_log_msg_cls(record.levelno)()
        if hasattr(record, "log_name"):
            msg.name = record.log_name  # type: ignore
        else:
            msg.name = record.name
        msg.time = record.created
        msg.level = record.levelno
        msg.lineno = record.lineno
        msg.pathname = record.pathname
        msg.funcname = record.funcName
        msg_str = record.getMessage()
        if len(msg_str) > (cd.MAX_LOG_LENGTH - 1):
            # truncate
            truncate_str = "[TRUNCATED]"
            trunc_len = cd.MAX_LOG_LENGTH - len(truncate_str) - 1
            msg_str = msg_str[:trunc_len] + truncate_str
        msg.message = msg_str
        return msg

    def emit(self, record: logging.LogRecord):
        client = self.client_ref()
        if client:
            try:
                if client.connected:
                    msg = self.gen_log_msg(record)
                    client.send_message(msg)
            except ClientError:
                if client:
                    client.logger.enable_rtma = False
            except Exception:
                if client:
                    client.logger.enable_rtma = False
                self.handleError(record)


class RichLogFormatter(logging.Formatter):
    """Custom Formatter that calls rich.markup.escape on record.msg so
    that rich formatting can be applied to formatting ONLY
    """

    def format(self, record: logging.LogRecord, *args):
        record.msg = escape(record.msg)
        return super().format(record, *args)


class RtmaLogFilter(logging.Filter):

    def __init__(self, log_name: str):
        self.log_name = log_name

    def filter(self, record: logging.LogRecord):
        record.log_name = self.log_name
        return True


class RTMALogger(object):
    def __init__(
        self,
        log_name: str,
        rtma_client: ClientLike,
        level: int = logging.NOTSET,
    ):
        # default formatter
        self._default_fmt = "{levelname:<8} - {asctime} - {log_name:<16} - {message}"
        self._rich_fmt = "[bold yellow]{log_name:<16}[/]   {message}"
        self._default_formatter: Optional[logging.Formatter] = logging.Formatter(
            self._default_fmt, style="{"
        )
        self._console_formatter: Optional[logging.Formatter] = RichLogFormatter(
            self._rich_fmt, style="{"
        )
        self._file_formatter: Optional[logging.Formatter] = self._default_formatter

        # initialize private attributes
        self._log_name = log_name
        self.children: List[logging.Logger] = []
        self._logger = logging.getLogger(hex(id(rtma_client)))
        self._logger.propagate = True
        self.set_all_levels(level)
        self._filter = RtmaLogFilter(self._log_name)
        self._logger.addFilter(self._filter)
        self._rtma_client_ref = weakref.ref(rtma_client)
        self._rtma_handler: Optional[RTMALogHandler] = self.init_rtma_handler()
        self._file_handler: Optional[logging.FileHandler] = None
        self._console_handler: Optional[logging.Handler] = self.init_console_handler()

        # default values for which handlers will be enabled
        if self._rtma_handler:
            self._logger.addHandler(self._rtma_handler)
        self._enable_rtma = self._rtma_handler is not None

        if self._console_handler:
            self._logger.addHandler(self._console_handler)
        self._enable_console = self._console_formatter is not None

        self._enable_file = False
        self._log_filename: Union[str, pathlib.Path] = ""

        # alias logger methods
        self.debug = self.logger.debug
        self.info = self.logger.info
        self.warning = self.logger.warning
        self.warn = self.logger.warn
        self.error = self.logger.error
        self.exception = self.logger.exception
        self.critical = self.logger.critical

    def __del__(self):
        if self._console_handler:
            self._logger.removeHandler(self._console_handler)

        if self._file_handler:
            self._logger.removeHandler(self._file_handler)
            self._file_handler = None

        if self._rtma_handler:
            self._logger.removeHandler(self._rtma_handler)

        # remove any other handlers
        for handler in self._logger.handlers:
            self._logger.removeHandler(handler)

    @property
    def logger(self) -> logging.Logger:  # read only
        return self._logger

    @property
    def log_name(self) -> str:  # read only
        return self._log_name  # self.logger.name #

    @log_name.setter
    def log_name(self, log_name: str):
        old_name = self._log_name
        self._log_name = log_name
        self._filter.log_name = log_name
        for child in self.children:
            for filter in child.filters:
                if isinstance(filter, RtmaLogFilter):
                    filter.log_name = filter.log_name.replace(old_name, log_name, 1)

    @property
    def console_handler(self) -> Optional[logging.Handler]:  # read only
        if self._enable_console:
            for handler in self._logger.handlers:
                if handler is self._console_handler:
                    return self._console_handler
        return None

    @property
    def rtma_handler(self) -> Optional[RTMALogHandler]:  # read only
        if self._enable_rtma:
            for handler in self._logger.handlers:
                if handler is self._rtma_handler:
                    return self._rtma_handler
        return None

    @property
    def file_handler(self) -> Optional[logging.FileHandler]:  # read only
        if self._enable_file and self._file_handler is not None:
            for handler in self._logger.handlers:
                if handler is self._file_handler:
                    return self._file_handler
        return None

    @property
    def enable_console(self) -> bool:
        return self._enable_console

    @enable_console.setter
    def enable_console(self, value: bool):
        if self._enable_console != bool(value):  # value changed
            if self.console_handler:
                if bool(value):
                    # add handler
                    self.logger.addHandler(self.console_handler)
                else:
                    # disable handler
                    self.logger.removeHandler(self.console_handler)
            self._enable_console = bool(value)

    @property
    def enable_rtma(self) -> bool:
        return self._enable_rtma

    @enable_rtma.setter
    def enable_rtma(self, value: bool):
        if self._enable_rtma != bool(value):  # value changed
            if self._rtma_handler:
                if bool(value):
                    # add handler
                    self.logger.addHandler(self._rtma_handler)
                else:
                    # disable handler
                    self.logger.removeHandler(self._rtma_handler)
            self._enable_rtma = bool(value)

    @property
    def enable_file(self) -> bool:
        return self._enable_file

    @enable_file.setter
    def enable_file(self, value: bool):
        if self._enable_file != bool(value):  # value changed
            if bool(value):
                if self._file_handler:
                    self.logger.removeHandler(self._file_handler)
                    self._file_handler = None

                # add handler
                self._file_handler = self.init_file_handler()
                self.logger.addHandler(self._file_handler)
            else:
                # disable handler
                if self._file_handler:
                    self.logger.removeHandler(self._file_handler)
                    self._file_handler = None

            self._enable_file = bool(value)

    @property
    def level(self) -> int:
        return self.logger.getEffectiveLevel()

    @level.setter
    def level(self, value: int):
        self._logger.setLevel(value)

    def add_child(self, child_name: str) -> logging.Logger:
        child_logger = self._logger.getChild(child_name)
        child_logger.setLevel(self.level)
        full_child_name = f"{self.log_name}.{child_name}"
        filter = RtmaLogFilter(full_child_name)
        child_logger.addFilter(filter)
        self.children.append(child_logger)
        return child_logger

    def set_all_levels(self, value: int):
        """Set the log level as well as the level for each handler"""
        # set main level
        self.level = value
        # set all handlers
        for handler in self._logger.handlers:
            handler.setLevel(value)

        # set private values
        self._console_level = value
        self._rtma_level = value
        self._file_level = value

        # set child levels
        for child in self.children:
            if value < child.getEffectiveLevel():
                child.setLevel(value)

    @property
    def log_filename(self) -> Union[str, pathlib.Path]:
        if self.file_handler:
            return self.file_handler.baseFilename
        else:
            return self._log_filename

    @log_filename.setter
    def log_filename(self, value: Union[str, pathlib.Path]):
        self._log_filename = value

    @property
    def console_level(self) -> int:
        if self.console_handler:
            self._console_level = self.console_handler.level
        return self._console_level

    @console_level.setter
    def console_level(self, value: int):
        if self.console_handler:
            self.console_handler.level = value
        self._console_level = value

    @property
    def rtma_level(self) -> int:
        if self.rtma_handler:
            self._rtma_level = self.rtma_handler.level
        return self._rtma_level

    @rtma_level.setter
    def rtma_level(self, value: int):
        if self.rtma_handler:
            self.rtma_handler.level = value
        self._rtma_level = value

    @property
    def file_level(self) -> int:
        if self.file_handler:
            self._file_level = self.file_handler.level
        return self._file_level

    @file_level.setter
    def file_level(self, value: int):
        if self.file_handler:
            self.file_handler.level = value
        self._file_level = value

    @property
    def console_formatter(self) -> Optional[logging.Formatter]:
        if self.console_handler:
            self._console_formatter = self.console_handler.formatter
        return self._console_formatter

    @console_formatter.setter
    def console_formatter(self, value: Optional[logging.Formatter]):
        if self.console_handler:
            self.console_handler.setFormatter(value)
        self._console_formatter = value

    @property
    def file_formatter(self) -> Optional[logging.Formatter]:
        if self.file_handler:
            self._file_formatter = self.file_handler.formatter
        return self._file_formatter

    @file_formatter.setter
    def file_formatter(self, value: Optional[logging.Formatter]):
        if self._file_handler:
            self._file_handler.setFormatter(value)
        self._file_formatter = value

    def init_console_handler(self) -> logging.Handler:
        # set markup to True if console formatter is RichLogFormatter
        # This will allow rich markup in the format string but NOT in each log message text (for compatibility with other handlers)
        console_handler = RichHandler(
            log_time_format="[%X.%f]",
            markup=isinstance(self._console_formatter, RichLogFormatter),
        )
        console_handler.name = "Console Handler"
        console_handler.setLevel(self._console_level)
        console_handler.setFormatter(self._console_formatter)

        return console_handler

    def init_rtma_handler(self):
        rtma_handler = RTMALogHandler(self._rtma_client_ref)
        rtma_handler.name = "RTMA Handler"
        rtma_handler.setLevel(self._rtma_level)

        return rtma_handler

    def init_file_handler(self) -> logging.FileHandler:
        file_handler = logging.handlers.RotatingFileHandler(
            self._log_filename, maxBytes=3 * (1024**2), backupCount=3
        )
        file_handler.name = "File Handler"
        file_handler.setLevel(self._file_level)
        file_handler.setFormatter(self._file_formatter)

        return file_handler

    @contextmanager
    def exception_logging_context(self):
        try:
            yield
        except Exception as e:
            e_str = str(e)
            e_type = type(e).__name__
            tb = traceback.extract_tb(e.__traceback__)
            tb_line = tb[1].lineno
            tb_fcn_name = tb[1].name
            tb_pathname = tb[1].filename
            tb_filename = pathlib.Path(tb_pathname).name
            if tb_fcn_name != "<module>":
                tb_fcn_name += "()"

            if e.__traceback__:
                exc_info = (type(e), e, e.__traceback__.tb_next)
            else:
                exc_info = (type(e), e, e.__traceback__)

            # there does not appear to be a (simple) way to point the log record to the line raising the exception
            # but exc_info prints it
            self.exception(
                f"{e_type}: {e_str} -- Unhandled exception in {tb_fcn_name} {tb_filename}:{tb_line}",
                exc_info=exc_info,
            )
            raise

    def __repr__(self):
        level = logging.getLevelName(self.logger.getEffectiveLevel())
        return "<%s %s (%s)>" % (self.__class__.__name__, self.log_name, level)



"""pyrtma.messaage: RTMA message classes"""

from __future__ import annotations

import json
import copy

from typing import Type, Dict, Any, TypeVar

from .header import MessageHeader, get_header_cls
from .message_data import MessageData
from .message_base import RTMAJSONEncoder
from .exceptions import InvalidMessageDefinition, UnknownMessageType

__all__ = [
    "Message",
    "MessageHeader",
    "MessageData",
    "get_header_cls",
    "get_msg_cls",
    "message_def",
    "msg_def",  # deprecated
]

# Main Map of all internal message types
_msg_defs: Dict[int, Type[MessageData]] = {}


_MD = TypeVar("_MD", bound=MessageData)  # Parent


def message_def(msg_cls: Type[_MD], *args, **kwargs) -> Type[_MD]:
    """Decorator to add user message definitions."""
    _msg_defs[msg_cls.type_id] = msg_cls
    return msg_cls


# backwards compatibility: deprecated name
msg_def = message_def


def _set_msg_defs(defs: Dict[int, Type[MessageData]]):
    _msg_defs.clear()
    _msg_defs.update(defs)


def _get_msg_defs() -> Dict[int, Type[MessageData]]:
    return copy.deepcopy(_msg_defs)


def _update_msg_defs(defs: Dict[int, Type[MessageData]]):
    _msg_defs.update(defs)


def _clear_msg_defs():
    _msg_defs.clear()


def get_msg_cls(id: int) -> Type[MessageData]:
    """get msg class for a given message type ID

    Args:
        id (int): Message Type ID

    Raises:
        UnknownMessageType: Message type is undefined

    Returns:
        Type[MessageData]: Message class
    """
    try:
        return _msg_defs[id]
    except KeyError as e:
        raise UnknownMessageType(
            f"There is no message definition associated with id: {id}"
        ) from e


class Message:
    """Message class

    Contains message header and data
    """

    def __init__(self, header: MessageHeader, data: MessageData):
        self.header = header
        self.data = data

    @property
    def type_id(self) -> int:
        return self.data.type_id

    @property
    def name(self) -> str:
        return self.data.type_name

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, Message):
            return False
        return self.header == other.header and self.data == other.data

    def pretty_print(self, add_tabs: int = 0) -> str:
        """Generate formatted string for pretty printing of message

        Args:
            add_tabs (int, optional): Indent level. Defaults to 0.

        Returns:
            str: Formatted string
        """
        return (
            self.header.pretty_print(add_tabs) + "\n" + self.data.pretty_print(add_tabs)
        )

    def __str__(self) -> str:
        return self.pretty_print()

    def to_dict(self) -> Dict[str, Any]:
        """Convert message to dictionary

        Returns:
            Dict[str, Any]: Message dictionary
        """
        return dict(header=self.header.to_dict(), data=self.data.to_dict())

    def to_json(self, minify: bool = False, **kwargs) -> str:
        """Convert message to JSON string

        Args:
            minify (bool, optional): Flag to minify (compact format). Defaults to False.

        Returns:
            str: JSON message string
        """
        d = dict(header=self.header.to_dict(), data=self.data.to_dict())
        if minify:
            return json.dumps(d, cls=RTMAJSONEncoder, separators=(",", ":"), **kwargs)
        else:
            return json.dumps(d, cls=RTMAJSONEncoder, indent=2, **kwargs)

    @classmethod
    def from_json(cls, s: str) -> Message:
        """Create message object from JSON string

        Args:
            s (str): JSON message string

        Raises:
            InvalidMessageDefinition: JSON data does not match expected message defintion

        Returns:
            Message: Message object
        """
        # Convert json string to dict
        d = json.loads(s)

        # Decode header segment
        hdr_cls = get_header_cls()
        hdr = hdr_cls.from_dict(d["header"])

        # Decode message data segment
        msg_cls = get_msg_cls(hdr.msg_type)

        # Note: Ignore the sync check if header.version is not filled in
        # This can removed once all clients support this field.
        if hdr.version != 0 and hdr.version != msg_cls.type_hash:
            raise InvalidMessageDefinition(
                f"Client's message definition does not match sender's version: {msg_cls.type_name}"
            )

        msg_data = msg_cls.from_dict(d["data"])

        obj = cls(hdr, msg_data)

        return obj

    @classmethod
    def copy(cls, m: Message) -> Message:
        """Generate a copy of a message structure

        Args:
            m: Message structure to copy
        """
        return Message(
            MessageHeader.from_buffer_copy(m.header), m.data.from_buffer_copy(m.data)
        )



"""pyrtma.manager module

Contains :py:class:`~MessageManager` class
"""

import socket
import select
import argparse
import logging
import time
import random
import ctypes
import os
import typing


from .client_logging import RTMALogger, ClientLike
from .validators import disable_message_validation
from .message import Message
from .header import MessageHeader, get_header_cls
from .message_data import MessageData
from .context import _get_core_defs
from .core_defs import ALL_MESSAGE_TYPES
from . import core_defs as cd

from typing import Dict, List, Tuple, Set, Type, Union, Optional
from itertools import chain
from dataclasses import dataclass, field
from collections import defaultdict, Counter
from contextlib import contextmanager
from contextvars import ContextVar


@dataclass
class Module:
    """Module dataclass

    Used internally by MessageManager to manage connections to each client module.
    """

    uid: int
    conn: socket.socket
    address: Tuple[str, int]
    header_cls: Type[MessageHeader]
    name: str = ""
    mod_id: int = 0
    pid: int = 0
    subs: Set[int] = field(default_factory=set)
    connected: bool = False
    is_logger: bool = False
    is_daemon: bool = False
    unique: bool = True
    drops: int = 0
    msg_count: int = 0

    @property
    def ipaddr(self) -> str:
        return f"{self.address[0]}:{self.address[1]}"

    @property
    def addr(self) -> str:
        return self.address[0]

    @property
    def port(self) -> int:
        return self.address[1]

    @property
    def sub_all(self) -> bool:
        return ALL_MESSAGE_TYPES in self.subs

    def send_message(self, header: MessageHeader, payload: Union[bytes, MessageData]):
        """Send a message

        Args:
            header (MessageHeader): Message header
            payload (Union[bytes, MessageData]): Message data
        """
        self.msg_count += 1
        header.msg_count = self.msg_count

        self.conn.sendall(header)
        self.conn.sendall(payload)

    def send_ack(self):
        """Send ACKNOWLEDGE signal header"""
        # Just send a header
        header = self.header_cls()
        header.msg_type = cd.MT_ACKNOWLEDGE
        header.send_time = time.perf_counter()
        header.src_mod_id = cd.MID_MESSAGE_MANAGER
        header.dest_mod_id = self.mod_id
        header.num_data_bytes = 0

        self.msg_count += 1
        header.msg_count = self.msg_count

        self.conn.sendall(header)

    def close(self):
        """Close connection"""
        self.conn.close()
        self.connected = False

    def __str__(self):
        return f"{self.name or 'ID'}({self.mod_id}) @ {self.ipaddr}"

    def __hash__(self):
        return self.conn.__hash__()


class MessageManager(ClientLike):
    """MessageManager class

    RTMA Message Manager server implemented in python.
    """

    TRAFFIC_INTERVAL = 1.0
    INFO_INTERVAL = 5.0

    def __init__(
        self,
        ip_address: str = "",  # "" equivalent to socket.INADDR_ANY
        port: int = 7111,
        timecode=False,
        log_level=logging.INFO,
        debug=False,
        send_msg_timing=True,
    ):
        """MessageManager class

        RTMA Message Manager server implemented in python.

        Args:
            ip_address (str, optional): server IP address. Defaults to "".
            port (int, optional): server port. Defaults to 7111.
            timecode (bool, optional): Flag to use message header with timecode values. Defaults to False.
            log_level (int, optional): logging level, defaults to logging.INFO.
            debug (bool, optional): Flag for debug mode. Defaults to False.
            send_msg_timing (bool, optional): Flag to send TIMING_MSG. Defaults to True.
        """
        self._keep_running = False
        self.ip_address = ip_address
        self.port = port

        self.header_cls = get_header_cls(timecode)
        self.header_size = ctypes.sizeof(self.header_cls)
        self.header_buffer = bytearray(self.header_size)
        self.header_view = memoryview(self.header_buffer)

        self.read_timeout = 0.200
        self.write_timeout = 0  # c++ message manager uses timeout = 0 for all modules except logger modules, which uses -1 (blocking)
        self._debug = debug
        self.b_send_msg_timing = send_msg_timing

        self._logger = RTMALogger(f"message_manager", self, logging.INFO)
        self.logger.set_all_levels(log_level)

        if ip_address == socket.INADDR_ANY:
            ip_address = ""  # bind and Module require a string input, '' is treated as INADDR_ANY by bind

        # Create the tcp listening socket
        self.listen_socket = socket.socket(
            family=socket.AF_INET, type=socket.SOCK_STREAM, proto=socket.IPPROTO_TCP
        )
        self.listen_socket.bind((ip_address, port))
        self.listen_socket.listen(socket.SOMAXCONN)
        self.modules: Dict[socket.socket, Module] = {}
        self.logger_modules: Set[Module] = set()
        self.next_dynamic_mod_id_offset = 0

        self.subscriptions: Dict[int, Set[Module]] = defaultdict(set)
        self.sockets = [self.listen_socket]
        self.start_time = time.time()

        # dictionary of message type ids and message counts, reset each time timing_message is sent
        self.message_counts: typing.Counter[int] = Counter()
        self.t_last_message_count = time.perf_counter()
        self.min_timing_message_period = 0.9

        self.last_client_info: float = time.perf_counter()
        self.sending_traffic: ContextVar[bool] = ContextVar(
            "sending_traffic", default=False
        )
        self.traffic_counter: typing.Counter[int] = Counter()
        self.traffic_start: float = time.perf_counter()
        self.traffic_seqno: int = 1

        # Disable Nagle Algorithm
        self.listen_socket.setsockopt(
            socket.getprotobyname("tcp"), socket.TCP_NODELAY, 1
        )

        self._uid = 0
        self.wlist: List[socket.socket] = []

        # Add message manager to its module list
        self.mm_module = Module(
            uid=0,
            conn=self.listen_socket,
            address=(ip_address, port),
            header_cls=self.header_cls,
            name="message_manager",
            mod_id=0,
            pid=os.getpid(),
            connected=True,
            is_logger=False,
        )

        self.modules[self.listen_socket] = self.mm_module

        self.data_buffer = bytearray(1024**2)
        self.data_view = memoryview(self.data_buffer)

        # Address Reuse allowed for testing
        if debug:
            self.listen_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

        self.logger.info("Message Manager Initialized.")

    @property
    def connected(self) -> bool:
        return True

    @property
    def logger(self) -> RTMALogger:
        return self._logger

    @contextmanager
    def sending_traffic_ctx(self):
        token = self.sending_traffic.set(True)
        yield
        self.sending_traffic.reset(token)

    def generate_uid(self) -> int:
        self._uid += 1
        return self._uid

    def assign_module_id(self) -> int:
        """Assign module ID dynamically to connecting module

        Raises:
            RuntimeError: Exceeded maximum number of allowed dynamic modules

        Returns:
            int: module ID
        """
        current_ids = [mod.mod_id for mod in self.modules.values()]

        MAX_DYN_IDS = cd.MAX_MODULES - cd.DYN_MOD_ID_START
        for i in range(0, MAX_DYN_IDS):
            mod_id = self.next_dynamic_mod_id_offset + cd.DYN_MOD_ID_START
            self.next_dynamic_mod_id_offset += 1
            if self.next_dynamic_mod_id_offset == MAX_DYN_IDS:
                self.next_dynamic_mod_id_offset = 0

            # check if mod id is already used, if it is, continue looping until we find an unused one
            if mod_id not in current_ids:
                return mod_id

        # if we exit loop without returning, we failed to find a valid id
        self.logger.error(
            f"MessageManager::assign_module_id: All valid dynamic IDs are in use"
        )

        raise RuntimeError("Exceeded maximum limit of allowed modules.")

    def decode_header(self) -> MessageHeader:
        return self.header_cls.from_buffer(self.header_view)

    def connect_module(self, module: Module, msg: Message) -> bool:
        """Connect module

        Args:
            module (Module): Connecting module
            msg (Message): Incoming connect message

        Returns:
            bool: success code
        """
        # ignore v1 message that follows v2 message
        if module.connected:
            return False

        if isinstance(msg.data, cd.MDF_CONNECT_V2):
            module.mod_id = msg.data.mod_id
            module.unique = msg.data.allow_multiple == 0
            module.pid = msg.data.pid
            module.name = msg.data.name
        elif isinstance(msg.data, cd.MDF_CONNECT):
            module.mod_id = msg.header.src_mod_id
        else:
            raise RuntimeError(
                f"MessageManager::connect_module: Uknown Connect message definition"
            )

        # fields common to v1 and v2
        module.is_logger = msg.data.logger_status == 1
        module.is_daemon = msg.data.daemon_status == 1

        if module.mod_id != 0:
            if module.mod_id < 1 or module.mod_id > cd.DYN_MOD_ID_START:
                self.logger.error(
                    f"Invalid Module id specified: {module.mod_id}. User assigned ids must be in range or 1 - {cd.DYN_MOD_ID_START}"
                )
                self.remove_module(module)
                return False

            for m in self.modules.values():
                if m is module:
                    continue

                if m.mod_id == module.mod_id:
                    if m.unique:
                        self.logger.error(
                            f"SET_ID - {module.ipaddr} - ID({module.mod_id}) - ID already in use. Closing connection."
                        )
                        self.remove_module(module)
                        return False

                    if module.unique:
                        self.logger.error(
                            f"SET_ID - {module.ipaddr} - ID({module.mod_id}) - ID already in use. Closing connection."
                        )
                        self.remove_module(module)
                        return False

                if module.name:
                    if (m.unique or module.unique) and (m.name == module.name):
                        self.logger.error(
                            f"SET_NAME - {module.ipaddr} - ID({module.mod_id}) - {module.name} - Name already in use."
                        )
                        self.remove_module(module)
                        return False

                    self.logger.debug(
                        f"SET_NAME - {module.ipaddr} - ID({module.mod_id}) - {module.name}"
                    )

        else:
            module.mod_id = self.assign_module_id()

        module.connected = True

        if module.is_logger:
            self.logger_modules.add(module)

        return True

    def remove_module(self, module: Module):
        """Remove connected module

        Args:
            module (Module): Module object to remove
        """
        # Drop all subscriptions for this module
        for msg_type in module.subs:
            self.subscriptions[msg_type].discard(module)

        # Discard from logger module set if needed
        self.logger_modules.discard(module)

        # Drop from our module mapping
        module.close()

        self.send_client_close(module)
        del self.modules[module.conn]

    def disconnect_module(self, src_module: Module):
        """Disconnect module

        Args:
            src_module (Module): Module object to disconnect
        """
        self.remove_module(src_module)

    def add_subscription(self, src_module: Module, msg: Message):
        """Add message subscription

        Args:
            src_module (Module): Subscribing module
            msg (Message): incoming SUBSCRIBE message
        """
        sub = cd.MDF_SUBSCRIBE.from_buffer(msg.data)

        if sub.msg_type == ALL_MESSAGE_TYPES:
            self.subscriptions[sub.msg_type].add(src_module)

            # Clear out the individual subs
            for sub_type in src_module.subs:
                self.subscriptions[sub_type].discard(src_module)
            src_module.subs.clear()

            src_module.subs.add(sub.msg_type)
            self.logger.debug(f"SUBSCRIBE- {src_module!s} to ALL_MESSAGE_TYPES")
        else:
            # Ignore individual msg_types subs when subscribed to ALL_MESSAGE_TYPES
            if src_module.sub_all:
                return
            self.subscriptions[sub.msg_type].add(src_module)
            src_module.subs.add(sub.msg_type)
            self.logger.debug(f"SUBSCRIBE- {src_module!s} to MT:{sub.msg_type}")

    def remove_subscription(self, src_module: Module, msg: Message):
        """Remove message subscription

        Args:
            src_module (Module): Unsubscribing module
            msg (Message): incoming UNSUBSCRIBE message
        """
        unsub = cd.MDF_UNSUBSCRIBE.from_buffer(msg.data)

        if unsub.msg_type == ALL_MESSAGE_TYPES:
            self.subscriptions[unsub.msg_type].discard(src_module)

            # Clear out the individual subs
            for sub_type in src_module.subs:
                self.subscriptions[sub_type].discard(src_module)
            src_module.subs.clear()

            self.logger.debug(f"UNSUBSCRIBE- {src_module!s} from ALL_MESSAGE_TYPES")
        else:
            # Ignore individual msg_types unsubs when subscribed to ALL_MESSAGE_TYPES
            if src_module.sub_all:
                return
            self.subscriptions[unsub.msg_type].discard(src_module)
            src_module.subs.discard(unsub.msg_type)
            self.logger.debug(f"UNSUBSCRIBE- {src_module!s} from MT:{unsub.msg_type}")

    def resume_subscription(self, src_module: Module, msg: Message):
        """Resume message subscription

        Args:
            src_module (Module): Subscribing module
            msg (Message): incoming RESUME_SUBSCRIPTION message
        """
        self.add_subscription(src_module, msg)

    def pause_subscription(self, src_module: Module, msg: Message):
        """Pause message subscription

        Args:
            src_module (Module): Subscribing module
            msg (Message): incoming PAUSE_SUBSCRIPTION message
        """
        self.remove_subscription(src_module, msg)

    def register_module_ready(self, src_module: Module, msg: Message):
        """Handle MODULE_READY message and register PID

        Args:
            src_module (Module): Module that is ready
            msg (Message): Incoming MODULE_READY message
        """
        mr = cd.MDF_MODULE_READY.from_buffer(msg.data)
        src_module.pid = mr.pid

    def set_module_name(self, src_module: Module, msg: Message):
        """Set a module name

        Args:
            src_module (Module): Module that sent CLIENT_SET_NAME
            msg (Message): Incoming CLIENT_SET_NAME message
        """
        name_msg = cd.MDF_CLIENT_SET_NAME.from_buffer(msg.data)
        src_module.name = name_msg.name or ""
        self.logger.info(
            f"SET_NAME - {src_module.ipaddr} - ID({src_module.mod_id}) - {src_module.name}"
        )

    def read_message(self, sock: socket.socket) -> Optional[MessageHeader]:
        """Read an incoming message

        Args:
            sock (socket.socket): socket to read from

        Returns:
            MessageHeader: message header of incoming message
        """
        # Read RTMA Header Section
        nbytes = sock.recv_into(
            self.header_buffer, self.header_size, socket.MSG_WAITALL
        )

        # Module that sent the message
        mod = self.modules[sock]

        if nbytes != self.header_size:
            self.remove_module(mod)
            self.logger.warning(
                f"DROPPING - {mod!s} - Wrong number of header bytes returned from sock.recv_into. Got ({nbytes})."
            )
            return

        header = self.decode_header()

        # Read Data Section into Internal Buffer
        data_size = header.num_data_bytes
        if data_size:
            if data_size > len(self.data_buffer):
                self.logger.warning(
                    "Message Data size (data_size) exceeds buffer size. Header may be corrupted."
                )
                self.remove_module(mod)
                return

            nbytes = sock.recv_into(self.data_buffer, data_size, socket.MSG_WAITALL)

            if nbytes != data_size:
                self.logger.warning(
                    f"DROPPING - {mod!s} - Wrong number of data bytes returned from sock.recv_into. Got ({nbytes})"
                )
                self.remove_module(mod)
                return

        return header

    def forward_message(
        self,
        src_module: Module,
        header: MessageHeader,
        data: Union[bytes, MessageData],
    ):
        """Forward a message from other modules

        The given message will be forwarded to:

            - all subscribed logger modules
            - if the message has a destination address, and it is subscribed to by that destination it will be forwarded only there
            - if the message has no destination address, it will be forwarded to all subscribed modules or those subscribed to ALL_MESSAGE_TYPES

        Args:
            src_module (Module): Module where message originated from
            header (MessageHeader): Message Header
            data (Union[bytes, MessageData]): Message Data
        """

        src_name = src_module.name or f"Module({src_module.mod_id})"

        # Increment message counts
        # Note: skip traffic count if we are currently forwarding out traffic messages)
        if not self.sending_traffic.get():
            if self.b_send_msg_timing:
                self.message_counts[header.msg_type] += 1
            self.traffic_counter[header.msg_type] += 1

        dest_mod_id = header.dest_mod_id
        dest_host_id = header.dest_host_id

        # Verify that the module & host ids are valid
        if dest_mod_id < 0 or dest_mod_id > cd.MAX_MODULES:
            self.logger.error(
                f"MessageManager::forward_message: Got invalid dest_mod_id [{dest_mod_id}] from {src_name}"
            )
            return

        if dest_host_id < 0 or dest_host_id > cd.MAX_HOSTS:
            self.logger.error(
                f"MessageManager::forward_message: Got invalid dest_host_id [{dest_host_id}] from {src_name}"
            )
            return

        # Subscriber set for this message type
        subscribers = list(
            chain(
                self.subscriptions[header.msg_type],
                self.subscriptions[ALL_MESSAGE_TYPES],
            )
        )

        for n in range(len(subscribers)):
            module = subscribers[n]
            if module.conn in self.wlist:
                try:
                    if (
                        dest_mod_id == 0
                        or (module.mod_id == dest_mod_id)
                        or module.is_logger
                    ):
                        module.send_message(header, data)
                        module.drops = 0
                except ConnectionError as err:
                    self.remove_module(module)
                    self.logger.error(
                        f"Connection Error on write to {module!s} - {err!s}"
                    )
                    print("x", end="", flush=True)
                    self.send_failed_message(module, header, time.perf_counter())
            elif module.is_logger:
                # Block until logger is ready
                select.select([], [module.conn], [], None)

                try:
                    module.send_message(header, data)
                    module.drops = 0
                except ConnectionError as err:
                    self.remove_module(module)
                    self.logger.error(
                        f"Connection Error on write to {module!s} - {err!s}"
                    )
                    print("x", end="", flush=True)
                    # this could result in infinite recursion,
                    # this is prevented by send_failed_message returning if
                    # failed message type is failed_message.
                    self.send_failed_message(module, header, time.perf_counter())

            else:
                module.drops += 1
                print("x", end="", flush=True)
                self.send_failed_message(module, header, time.perf_counter())

    def send_to_loggers(
        self,
        header: MessageHeader,
        payload: Union[bytes, MessageData],
    ):
        """Forward message to registered logger modules

        Args:
            header (MessageHeader): Message header to send
            payload (Union[bytes, MessageData]): Message data to send
        """
        for module in self.logger_modules:
            if module.conn not in self.wlist:
                # Block until logger is ready
                select.select([], [module.conn], [], None)
            try:
                module.send_message(header, payload)
                module.drops = 0
            except ConnectionError as err:
                self.remove_module(module)
                self.logger.error(f"Connection Error on write to {module!s} - {err!s}")
                print("x", end="", flush=True)
                # this could result in infinite recursion,
                # this is prevented by send_failed_message returning if
                # failed message type is failed_message.
                self.send_failed_message(module, header, time.perf_counter())

    def send_message(
        self,
        msg_data: MessageData,
        dest_mod_id: int = 0,
        dest_host_id: int = 0,
        timeout: float = 0,
    ):
        """Send a message originating from MessageManager module

        Args:
            msg_data (MessageData): Object containing the message to send
            dest_mod_id (int, optional): Specific module ID to send to. Defaults to 0 (broadcast).
            dest_host_id (int, optional): Specific host ID to send to. Defaults to 0 (broadcast).
            timeout (float, optional): Reserved. Defaults to 0.
        """
        header = self.header_cls()
        header.msg_type = msg_data.type_id
        header.send_time = time.perf_counter()
        header.src_mod_id = cd.MID_MESSAGE_MANAGER
        header.dest_mod_id = dest_mod_id
        header.dest_host_id = dest_host_id
        header.num_data_bytes = msg_data.type_size

        self.forward_message(self.mm_module, header, msg_data)

    def send_ack(self, src_module: Module):
        """Send ACKNOWLEDGE signal header

        Args:
            src_module (Module): Module to send ACK to
        """
        header = self.header_cls()
        header.msg_type = cd.MT_ACKNOWLEDGE
        header.send_time = time.perf_counter()
        header.src_mod_id = cd.MID_MESSAGE_MANAGER
        header.dest_mod_id = src_module.mod_id
        header.num_data_bytes = 0

        try:
            src_module.send_message(header, b"")
        except ConnectionError as err:
            self.remove_module(src_module)
            self.logger.error(f"Connection Error on write to {src_module!s} - {err!s}")
            print("x", end="", flush=True)
            self.send_failed_message(src_module, header, time.perf_counter())

        # Always forward to logger modules
        self.send_to_loggers(header, b"")

    def send_failed_message(
        self,
        dest_module: Module,
        header: MessageHeader,
        time_of_failure: float,
    ):
        """Send FAILED_MESSAGE

        Args:
            dest_module (Module): Intended destination
            header (MessageHeader): Header of failed message
            time_of_failure (float): Time of send failure
        """
        # Avoid infinite recursion
        if header.msg_type in (
            cd.MT_FAILED_MESSAGE,
            cd.MT_RTMA_LOG,
            cd.MT_RTMA_LOG_CRITICAL,
            cd.MT_RTMA_LOG_ERROR,
            cd.MT_RTMA_LOG_WARNING,
            cd.MT_RTMA_LOG_INFO,
            cd.MT_RTMA_LOG_DEBUG,
        ):
            return

        out_header = self.header_cls()
        data = cd.MDF_FAILED_MESSAGE()
        out_header.msg_type = cd.MT_FAILED_MESSAGE
        out_header.send_time = time.perf_counter()
        out_header.src_mod_id = cd.MID_MESSAGE_MANAGER
        out_header.num_data_bytes = ctypes.sizeof(data)

        data.dest_mod_id = dest_module.mod_id
        data.time_of_failure = time_of_failure

        # Copy the values into the RTMA_MSG_HEADER
        for fname, ftype, *_ in data.msg_header._fields_:
            setattr(data.msg_header, fname, getattr(header, fname))

        # send to logger modules AND modules subscribed to FAILED_MESSAGE
        self.forward_message(self.mm_module, out_header, data)

    def send_timing_message(self):
        """Send TIMING_MESSAGE"""
        data = cd.MDF_TIMING_MESSAGE()
        for mt, count in self.message_counts.items():
            data.timing[mt] = count
        self.message_counts.clear()

        for mod in self.modules.values():
            data.ModulePID[mod.mod_id] = mod.pid

        data.send_time = time.perf_counter()
        with self.sending_traffic_ctx():
            self.send_message(data)

    def send_traffic(self):
        """Send MESSAGE_TRAFFIC"""
        with self.sending_traffic_ctx():
            self.logger.debug("MESSAGE_TRAFFIC")
            data = cd.MDF_MESSAGE_TRAFFIC()
            now = time.perf_counter()
            sub_seqno = 1
            nsent = 0
            i = -1
            for n, (mt, count) in enumerate(self.traffic_counter.items()):
                data.seqno = self.traffic_seqno
                data.sub_seqno = sub_seqno
                data.start_timestamp = self.traffic_start
                data.end_timestamp = now

                i = n % cd.MESSAGE_TRAFFIC_SIZE
                data.msg_type[i] = mt
                data.msg_count[i] = count

                if (n % cd.MESSAGE_TRAFFIC_SIZE) == 0:
                    nsent = n
                    self.send_message(data)
                    sub_seqno += 1

            # Send any remaining
            if i >= 0:
                i += 1
                if nsent < len(self.traffic_counter):
                    data.msg_type[i:] = [-1 for _ in range(cd.MESSAGE_TRAFFIC_SIZE - i)]
                    self.send_message(data)

        self.traffic_counter.clear()
        self.traffic_start = now
        self.traffic_seqno += 1

    def send_client_close(self, module: Module):
        """Send CLIENT_CLOSED

        Args:
            module (Module): Closed module object
        """
        self.logger.debug("CLIENT_CLOSE")
        msg = cd.MDF_CLIENT_CLOSED()
        msg.uid = module.uid
        msg.pid = module.pid
        msg.mod_id = module.mod_id
        msg.is_logger = module.is_logger
        msg.is_unique = module.unique
        msg.port = module.port
        msg.name = module.name
        msg.addr = module.addr
        self.send_message(msg)

    def send_client_info(self, module: Module):
        """Send CLIENT_INFO

        Args:
            module (Module): Module object to send info for
        """
        self.logger.debug("CLIENT_INFO")
        msg = cd.MDF_CLIENT_INFO()
        msg.uid = module.uid
        msg.pid = module.pid
        msg.mod_id = module.mod_id
        msg.is_logger = module.is_logger
        msg.is_unique = module.unique
        msg.port = module.port
        msg.name = module.name
        msg.addr = module.addr
        self.send_message(msg)

    def send_active_clients(self):
        """Send ACTIVE_CLIENTS"""
        self.logger.debug("ACTIVE_CLIENTS")
        msg = cd.MDF_ACTIVE_CLIENTS()
        msg.timestamp = time.perf_counter()

        for i, (sock, module) in enumerate(self.modules.items()):
            # if sock == self.listen_socket:
            #     continue
            msg.client_mod_id[i] = module.mod_id
            msg.client_pid[i] = module.pid
            self.send_client_info(module)

        msg.num_clients = len(self.modules) - 1
        self.send_message(msg)
        self.last_client_info = msg.timestamp

    def decode_core_message(self, hdr: MessageHeader) -> Message:
        data_cls = _get_core_defs().get(hdr.msg_type)
        if data_cls:
            data = data_cls.from_buffer(self.data_buffer)
            return Message(hdr, data)
        else:
            raise RuntimeError(f"Unknown core_def MT={hdr.msg_type}")

    def process_core_message(self, src_module: Module, header: MessageHeader):
        """Process incoming core message

        Args:
            src_module (Module): Message source module
            header (MessageHeader): Message header of the incoming message
        """
        msg_type = header.msg_type
        if msg_type >= 100:
            return

        core_msg = self.decode_core_message(header)

        if msg_type == cd.MT_CONNECT or msg_type == cd.MT_CONNECT_V2:
            if self.connect_module(src_module, core_msg):
                self.send_ack(src_module)
                self.send_client_info(src_module)
                if msg_type == cd.MT_CONNECT:
                    self.logger.info(f"CONNECT - {src_module!s}")
                else:
                    self.logger.info(f"CONNECT v2 - {src_module!s}")
        elif msg_type == cd.MT_DISCONNECT:
            self.disconnect_module(src_module)
            self.logger.info(f"DISCONNECT - {src_module!s}")
        elif msg_type == cd.MT_SUBSCRIBE:
            self.add_subscription(src_module, core_msg)
            self.send_ack(src_module)
        elif msg_type == cd.MT_UNSUBSCRIBE:
            self.remove_subscription(src_module, core_msg)
            self.send_ack(src_module)
        elif msg_type == cd.MT_PAUSE_SUBSCRIPTION:
            self.pause_subscription(src_module, core_msg)
            self.send_ack(src_module)
        elif msg_type == cd.MT_RESUME_SUBSCRIPTION:
            self.resume_subscription(src_module, core_msg)
            self.send_ack(src_module)
        elif msg_type == cd.MT_CLIENT_SET_NAME:
            self.set_module_name(src_module, core_msg)
            self.send_client_info(src_module)
        elif msg_type == cd.MT_MODULE_READY:
            self.register_module_ready(src_module, core_msg)
            self.send_client_info(src_module)

    def process_message(self, src_module: Module, header: MessageHeader):
        """Process incoming message

        Args:
            src_module (Module): Message source module
            header (MessageHeader): Message header of the incoming message
        """

        # Handle any internal core messages that come through
        self.process_core_message(src_module, header)

        # Forward all messages that come through to MM
        self.logger.debug(f"FORWARD - msg_type:{header.msg_type} from {src_module!s}")
        data = self.data_view[: header.num_data_bytes]
        self.forward_message(src_module, header, data)

    def close(self):
        """Close manager server"""
        self._keep_running = False

    def run(self):
        """Start the message manager server"""
        self._keep_running = True
        try:
            with disable_message_validation():
                while self._keep_running:
                    rlist, _, _ = select.select(
                        self.modules.keys(), [], [], self.read_timeout
                    )

                    # Check for an incoming connection request
                    if len(rlist) > 0:
                        try:
                            rlist.remove(self.listen_socket)
                            (conn, address) = self.listen_socket.accept()
                            self.logger.info(
                                f"New connection accepted from {address[0]}:{address[1]}"
                            )

                            # Disable Nagle Algorithm
                            conn.setsockopt(
                                socket.getprotobyname("tcp"), socket.TCP_NODELAY, 1
                            )

                            self.sockets.append(conn)
                            self.modules[conn] = Module(
                                self.generate_uid(), conn, address, self.header_cls
                            )
                        except ValueError:
                            pass

                        # Randomly select the order of sockets with data.
                        random.shuffle(rlist)

                        # Check whichs clients are ready to receive data
                        self.wlist.clear()
                        if rlist:
                            _, self.wlist, _ = select.select(
                                [], self.modules.keys(), [], self.write_timeout
                            )

                        for client_socket in rlist:
                            # Check that module is still active
                            src = self.modules.get(client_socket)
                            if src:
                                try:
                                    msg_hdr = self.read_message(client_socket)
                                except ConnectionError as err:
                                    self.disconnect_module(src)
                                    self.logger.error(
                                        f"Connection Error on read, disconnecting  {src!s} - {err!s}"
                                    )
                                    continue

                                if msg_hdr:
                                    self.process_message(src, msg_hdr)

                    now = time.perf_counter()

                    if (
                        self.b_send_msg_timing
                        and (now - self.t_last_message_count)
                        > self.min_timing_message_period
                    ):
                        self.send_timing_message()
                        self.t_last_message_count = now

                    if (now - self.traffic_start) > self.TRAFFIC_INTERVAL:
                        self.send_traffic()

                    if (now - self.last_client_info) > self.INFO_INTERVAL:
                        self.send_active_clients()

        except KeyboardInterrupt:
            self.logger.info("Stopping Message Manager")
        finally:
            for mod in self.modules:
                mod.close()


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-a",
        "--addr",
        type=str,
        default="",
        help="Listener address. IP address/hostname as a string. Default is '' which is evaluated as socket.INADDR_ANY.",
    )
    parser.add_argument(
        "-p", "--port", type=int, default=7111, help="Listener port. Default is 7111."
    )
    parser.add_argument("-d", "--debug", action="store_true", help="Debug mode")
    parser.add_argument(
        "--log-level",
        dest="log_level",
        choices=["DEBUG", "INFO", "WARN", "ERROR"],
        default="INFO",
        help="Logging Level",
    )
    parser.add_argument(
        "-t", "--timecode", action="store_true", help="Use timecode in message header"
    )
    parser.add_argument(
        "-T",
        "--disable_timing_msg",
        action="store_true",
        help="Disable sending of TIMING_MESSAGE",
    )
    args = parser.parse_args()

    if args.addr:  # a non-empty host address was passed in.
        ip_addr = args.addr
    else:
        ip_addr = ""  # socket.INADDR_ANY

    if args.log_level == "DEBUG":
        level = logging.DEBUG
    elif args.log_level == "INFO":
        level = logging.INFO
    elif args.log_level == "WARN":
        level = logging.WARN
    elif args.log_level == "ERROR":
        level = logging.ERROR
    else:
        print("Unknown log level. Using INFO instead")
        level = logging.INFO

    with disable_message_validation():
        msg_mgr = MessageManager(
            ip_address=ip_addr,
            port=args.port,
            timecode=args.timecode,
            log_level=level,
            debug=args.debug,
            send_msg_timing=(not args.disable_timing_msg),
        )

        msg_mgr.run()


if __name__ == "__main__":
    main()



import pyrtma
import pyrtma.core_defs as cd
import json

from pyrtma.core_defs import ALL_MESSAGE_TYPES


def add_collection(mod: pyrtma.Client):
    msg = cd.MDF_ADD_DATA_COLLECTION()
    msg.collection.name = input("(add-collection)->name: ")
    msg.collection.base_path = input("(add-collection)->base_path: ")
    msg.collection.dir_fmt = input("(add-collection)->dir_fmt: ")
    mod.send_message(msg)


def rm_collection(mod: pyrtma.Client):
    msg = cd.MDF_REMOVE_DATA_COLLECTION()
    msg.collection_name = input("(rm-collection)->name: ")
    mod.send_message(msg)


def add_data_set(mod: pyrtma.Client):
    msg = cd.MDF_ADD_DATA_SET()
    msg.collection_name = input("(add-data-set)->collection_name: ")
    msg.data_set.name = input("(add-data-set)->name: ")
    msg.data_set.sub_dir_fmt = input("(add-data-set)->sub_dir_fmt: ")
    msg.data_set.file_name_fmt = input("(add-data-set)->file_name_fmt: ")
    msg.data_set.formatter = input("(add-data-set)->formatter: ")

    raw = input("(add-data-set)->msg_type: ")
    msg_types = list(map(int, raw.split()))

    msg.data_set.msg_types[: len(msg_types)] = msg_types
    mod.send_message(msg)


def rm_data_set(mod: pyrtma.Client):
    msg = cd.MDF_REMOVE_DATA_SET()
    msg.collection_name = input("(rm-data-set)->collection_name: ")
    msg.name = input("(rm-data-set)->name: ")
    mod.send_message(msg)


def update_metadata(mod: pyrtma.Client, metadata):
    raw = input("(update-metadata)-> ")
    msg = cd.MDF_DATA_LOGGER_METADATA_UPDATE()
    msg.json = raw
    mod.send_message(msg)
    mod.send_signal(cd.MT_DATA_LOGGER_METADATA_REQUEST)


def main():
    import sys

    if len(sys.argv) > 1:
        server = sys.argv[1]
    else:
        server = "127.0.0.1:7111"

    mod = pyrtma.Client()
    mod.connect(server)
    mod.subscribe(
        [
            cd.MT_DATA_LOGGER_STATUS,
            cd.MT_DATA_COLLECTION_CONFIG,
            cd.MT_DATA_LOGGER_METADATA,
            cd.MT_DATA_LOGGER_ERROR,
        ]
    )

    metadata = {}

    try:
        while True:
            cmd = input("(data_log)>> ")

            if cmd == "add-collection":
                add_collection(mod)
            elif cmd == "add-data-set":
                add_data_set(mod)
            elif cmd == "rm-collection":
                rm_collection(mod)
            elif cmd == "rm-data-set":
                rm_data_set(mod)
            elif cmd == "pause":
                mod.send_signal(cd.MT_DATA_LOGGER_PAUSE)
            elif cmd == "resume":
                mod.send_signal(cd.MT_DATA_LOGGER_RESUME)
            elif cmd == "reset":
                mod.send_signal(cd.MT_DATA_LOGGER_RESET)
            elif cmd == "start":
                mod.send_signal(cd.MT_DATA_LOGGER_START)
            elif cmd == "stop":
                mod.send_signal(cd.MT_DATA_LOGGER_STOP)
            elif cmd == "get-status":
                mod.send_signal(cd.MT_DATA_LOGGER_STATUS_REQUEST)
            elif cmd == "get-config":
                mod.send_signal(cd.MT_DATA_COLLECTION_CONFIG_REQUEST)
            elif cmd == "get-metadata":
                mod.send_signal(cd.MT_DATA_LOGGER_METADATA_REQUEST)
            elif cmd == "update-metadata":
                update_metadata(mod, metadata)
            elif cmd == "help":
                help()
            elif cmd == "exit":
                break
            else:
                print(f"Unknown command: {cmd}")
                help()

            msg = mod.read_message(0.200)
            if msg:
                if isinstance(msg.data, cd.MDF_DATA_LOGGER_ERROR):
                    print(msg.data.msg)
                elif isinstance(msg.data, cd.MDF_DATA_LOGGER_STATUS):
                    print(msg.data.to_json())
                elif isinstance(msg.data, cd.MDF_DATA_LOGGER_METADATA):
                    print(msg.data.json)
                    metadata = json.loads(msg.data.json)
                elif isinstance(msg.data, cd.MDF_DATA_COLLECTION_CONFIG):
                    d = msg.data.collection.to_dict()
                    for ds in d["data_sets"]:
                        names = []
                        for msg_type in ds["msg_types"]:
                            if msg_type < 1:
                                continue
                            if msg_type == ALL_MESSAGE_TYPES:
                                names.append("ALL_MESSAGE_TYPES")
                            else:
                                names.append(pyrtma.get_msg_cls(msg_type).type_name)
                        ds["msg_types"] = names
                    d["data_sets"] = d["data_sets"][: d["num_data_sets"]]
                    print(json.dumps(d, indent=2))

    except KeyboardInterrupt:
        pass

    print("Goodbye")


def help():
    print("data_logger_control:")
    print("  * add-collection")
    print("  * add-data-set")
    print("  * rm-collection")
    print("  * rm-data-set")
    print("  * start")
    print("  * stop")
    print("  * pause")
    print("  * resume")
    print("  * reset")
    print("  * get-status")
    print("  * get-config")
    print("  * get-metadata")
    print("  * update-metadata")
    print("  * exit - close application.")
    print()


if __name__ == "__main__":
    main()



from .message import *
from .client import *
from .client_logging import RTMALogger as RTMALogger
from .context import *

from .__version__ import (
    __author__,
    __author_email__,
    __copyright__,
    __description__,
    __license__,
    __title__,
    __url__,
    __version__,
)



import pyrtma

from ..data_formatter import DataFormatter
from typing import ClassVar, IO
from itertools import chain


class MsgHeaderFormatter(DataFormatter):
    name: ClassVar[str] = "msg_header"
    ext: ClassVar[str] = ".csv"

    def __init__(self, fd: IO[str]):
        super().__init__(fd)

    def format_header(self) -> str:
        hdr_cls = pyrtma.get_header_cls()
        hdr = hdr_cls().to_dict()
        return ",".join(hdr.keys()) + "\n"

    def format_footer(self) -> str:
        return ""

    def format_message(self, msg: pyrtma.Message) -> str:
        d = msg.header.to_dict()
        return ",".join(chain(map(str, d.values()), ("\n",)))



import random
import threading
import time
import unittest
import logging

import pyrtma
from pyrtma import message_def
from pyrtma.message_base import MessageMeta
from pyrtma.client import Client, client_context
from pyrtma.manager import MessageManager
from pyrtma.validators import (
    Int32,
    Double,
    IntArray,
    ByteArray,
)

# Choose a unique message type id number
MT_TEST_MESSAGE = 123
MT_TEST_MESSAGE2 = 456


@message_def
class TEST_MESSAGE(pyrtma.MessageData, metaclass=MessageMeta):
    type_id: int = MT_TEST_MESSAGE
    type_name: str = "TEST_MESSAGE"
    type_size = 104
    type_source = ""
    type_def = ""
    type_hash = int("e3b0c442", 16)
    str = ByteArray(64)
    val = Double()
    arr = IntArray(Int32, 8)


@message_def
class TEST_MESSAGE2(pyrtma.MessageData, metaclass=MessageMeta):
    type_id: int = MT_TEST_MESSAGE2
    type_name: str = "TEST_MESSAGE2"
    type_size = 8
    type_source = ""
    type_def = ""
    type_hash = int("e3b0c442", 16)
    val = Double()


def wait_for_message():
    """
    Helper function for allowing time for a message to reach the manager.
    """
    time.sleep(0.1)


class TestSingleClient(unittest.TestCase):
    """
    Test interactions between a single client and manager.
    """

    def setUp(self):
        self.port = random.randint(1000, 10000)  # random port
        self.addr = f"127.0.0.1:{self.port}"
        self.module_id = 11
        self.client = Client(module_id=self.module_id, host_id=0, timecode=False)

        self.manager = MessageManager(
            ip_address="127.0.0.1",
            port=self.port,
            timecode=False,
            log_level=logging.ERROR,
            debug=False,
            send_msg_timing=True,
        )
        self.manager_thread = threading.Thread(
            target=self.manager.run,
        )
        self.manager_thread.start()
        wait_for_message()

    def tearDown(self):
        self.client.disconnect()
        self.manager.close()
        self.manager_thread.join()

    def test_whenClientConnects_clientConnectsToMessageManager(self):
        """
        Test if client module is connected when connect is called.
        """
        # Arrange

        # Act
        self.client.connect(server_name=self.addr)
        wait_for_message()

        # Assert
        self.assertTrue(self.client.connected, msg="Client not connected.")
        self.assertIn(
            self.module_id,
            [mod.mod_id for mod in self.manager.modules.values()],
            msg="Module id not found in Messager Manager modules.",
        )

    def test_whenClientSubscribes_clientIsSubscribedToMessageType(self):
        """
        Test if client is subscribed to message type when subscribe is called.
        """
        # Arrange
        self.client.connect(server_name=self.addr)
        wait_for_message()

        # Act
        self.client.subscribe([MT_TEST_MESSAGE])
        wait_for_message()

        # Assert
        self.assertIn(
            self.module_id,
            [mod.mod_id for mod in self.manager.subscriptions[MT_TEST_MESSAGE]],
            msg="Module id not found in TEST_MESSAGE subscriptions.",
        )

    def test_whenClientSubscribesMulti_clientIsSubscribedToMessageTypes(self):
        """
        Test if client is subscribed to message types when multiple message
            types are sent to subscribe.
        """
        # Arrange
        self.client.connect(server_name=self.addr)
        wait_for_message()

        # Act
        self.client.subscribe([MT_TEST_MESSAGE, MT_TEST_MESSAGE2])
        wait_for_message()

        # Assert
        self.assertIn(
            self.module_id,
            [mod.mod_id for mod in self.manager.subscriptions[MT_TEST_MESSAGE]],
            msg="Module id not found in TEST_MESSAGE subscriptions",
        )
        self.assertIn(
            self.module_id,
            [mod.mod_id for mod in self.manager.subscriptions[MT_TEST_MESSAGE2]],
            msg="Module id not found in TEST_MESSAGE2 subscriptions",
        )

    def test_whenClientUnsubscribes_clientIsUnsubscribedFromMessageType(self):
        """
        Test if client is unsubscribed from message type when unsubscribe is called.
        """
        # Arrange
        self.client.connect(server_name=self.addr)
        wait_for_message()
        self.client.subscribe([MT_TEST_MESSAGE])
        wait_for_message()

        # Act
        self.client.unsubscribe([MT_TEST_MESSAGE])
        wait_for_message()

        # Assert
        self.assertNotIn(
            self.module_id,
            [mod.mod_id for mod in self.manager.subscriptions[MT_TEST_MESSAGE]],
            msg="Module id found in TEST_MESSAGE subscriptions.",
        )

    def test_whenClientUnsubscribesMulti_clientIsUnsubscribedFromMessageTypes(self):
        """
        Test if client is unsubscribed from message types when multiple message
            types are sent to unsubscribe.
        """
        # Arrange
        self.client.connect(server_name=self.addr)
        wait_for_message()
        self.client.subscribe([MT_TEST_MESSAGE, MT_TEST_MESSAGE2])
        wait_for_message()

        # Act
        self.client.unsubscribe([MT_TEST_MESSAGE, MT_TEST_MESSAGE2])
        wait_for_message()

        # Assert
        self.assertNotIn(
            self.module_id,
            [mod.mod_id for mod in self.manager.subscriptions[MT_TEST_MESSAGE]],
            msg="Module id found in TEST_MESSAGE subscriptions.",
        )
        self.assertNotIn(
            self.module_id,
            [mod.mod_id for mod in self.manager.subscriptions[MT_TEST_MESSAGE2]],
            msg="Module id found in MT_TEST_MESSAGE2 subscriptions.",
        )

    def test_whenClientPauses_clientIsRemovedFromMessageType(self):
        """
        Test if client is unsubscribed from message type when pause is called.
        """
        # Arrange
        self.client.connect(server_name=self.addr)
        wait_for_message()
        self.client.subscribe([MT_TEST_MESSAGE])
        wait_for_message()

        # Act
        self.client.pause_subscription([MT_TEST_MESSAGE])
        wait_for_message()

        # Assert
        self.assertNotIn(
            self.module_id,
            [mod.mod_id for mod in self.manager.subscriptions[MT_TEST_MESSAGE]],
            msg="Module id found in TEST_MESSAGE subscriptions.",
        )

    def test_whenClientPausesMulti_clientIsRemovedFromMessageTypes(self):
        """
        Test if client is unsubscribed from message types when multiple message
            types are sent to pause.
        """
        # Arrange
        self.client.connect(server_name=self.addr)
        wait_for_message()
        self.client.subscribe([MT_TEST_MESSAGE, MT_TEST_MESSAGE2])
        wait_for_message()

        # Act
        self.client.pause_subscription([MT_TEST_MESSAGE, MT_TEST_MESSAGE2])
        wait_for_message()

        # Assert
        self.assertNotIn(
            self.module_id,
            [mod.mod_id for mod in self.manager.subscriptions[MT_TEST_MESSAGE]],
            msg="Module id found in TEST_MESSAGE subscriptions.",
        )
        self.assertNotIn(
            self.module_id,
            [mod.mod_id for mod in self.manager.subscriptions[MT_TEST_MESSAGE2]],
            msg="Module id found in MT_TEST_MESSAGE2 subscriptions.",
        )

    def test_whenClientResumes_clientIsAddedToMessageType(self):
        """
        Test if client is subscribed to message type when resume is called.
        """
        # Arrange
        self.client.connect(server_name=self.addr)
        wait_for_message()
        self.client.subscribe([MT_TEST_MESSAGE])
        wait_for_message()
        self.client.pause_subscription([MT_TEST_MESSAGE])
        wait_for_message()

        # Act
        self.client.resume_subscription([MT_TEST_MESSAGE])
        wait_for_message()

        # Assert
        self.assertIn(
            self.module_id,
            [mod.mod_id for mod in self.manager.subscriptions[MT_TEST_MESSAGE]],
            msg="Module id not found in TEST_MESSAGE subscriptions",
        )

    def test_whenClientResumesMulti_clientIsAddedToMessageType(self):
        """
        Test if client is subscribed to message types when multiple message
            types are sent to resume.
        """
        # Arrange
        self.client.connect(server_name=self.addr)
        wait_for_message()
        self.client.subscribe([MT_TEST_MESSAGE, MT_TEST_MESSAGE2])
        wait_for_message()
        self.client.pause_subscription([MT_TEST_MESSAGE, MT_TEST_MESSAGE2])
        wait_for_message()

        # Act
        self.client.resume_subscription([MT_TEST_MESSAGE, MT_TEST_MESSAGE2])
        wait_for_message()

        # Assert
        self.assertIn(
            self.module_id,
            [mod.mod_id for mod in self.manager.subscriptions[MT_TEST_MESSAGE]],
            msg="Module id not found in TEST_MESSAGE subscriptions",
        )
        self.assertIn(
            self.module_id,
            [mod.mod_id for mod in self.manager.subscriptions[MT_TEST_MESSAGE2]],
            msg="Module id not found in TEST_MESSAGE2 subscriptions",
        )

    def test_whenClientContextSubscribes_clientIsSubscribedToMessageType(self):
        """
        Test if client is subscribed to message type when subscribe is called.
        """
        # Arrange
        self.client.connect(server_name=self.addr)
        wait_for_message()

        # Act
        with self.client.subscription_context([MT_TEST_MESSAGE]):
            wait_for_message()

            # Assert
            self.assertIn(
                self.module_id,
                [mod.mod_id for mod in self.manager.subscriptions[MT_TEST_MESSAGE]],
                msg="Module id not found in TEST_MESSAGE subscriptions.",
            )

    def test_whenClientContextSubscribesMulti_clientIsSubscribedToMessageTypes(self):
        """
        Test if client is subscribed to message types when multiple message
            types are sent to subscribe.
        """
        # Arrange
        self.client.connect(server_name=self.addr)
        wait_for_message()

        # Act
        with self.client.subscription_context([MT_TEST_MESSAGE, MT_TEST_MESSAGE2]):
            wait_for_message()

            # Assert
            self.assertIn(
                self.module_id,
                [mod.mod_id for mod in self.manager.subscriptions[MT_TEST_MESSAGE]],
                msg="Module id not found in TEST_MESSAGE subscriptions",
            )
            self.assertIn(
                self.module_id,
                [mod.mod_id for mod in self.manager.subscriptions[MT_TEST_MESSAGE2]],
                msg="Module id not found in TEST_MESSAGE2 subscriptions",
            )

    def test_whenClientContextUnsubscribes_clientIsUnsubscribedFromMessageType(self):
        """
        Test if client is unsubscribed from message type when unsubscribe is called.
        """
        # Arrange
        self.client.connect(server_name=self.addr)
        wait_for_message()
        with self.client.subscription_context([MT_TEST_MESSAGE]):
            wait_for_message()

        # Act
        # unsubscribes by exiting context above
        wait_for_message()

        # Assert
        self.assertNotIn(
            self.module_id,
            [mod.mod_id for mod in self.manager.subscriptions[MT_TEST_MESSAGE]],
            msg="Module id found in TEST_MESSAGE subscriptions.",
        )

    def test_whenClientContextUnsubscribesMulti_clientIsUnsubscribedFromMessageTypes(
        self,
    ):
        """
        Test if client is unsubscribed from message types when multiple message
            types are sent to unsubscribe.
        """
        # Arrange
        self.client.connect(server_name=self.addr)
        wait_for_message()
        with self.client.subscription_context([MT_TEST_MESSAGE, MT_TEST_MESSAGE2]):
            wait_for_message()

        # Act
        # unsubscribes by exiting context above
        wait_for_message()

        # Assert
        self.assertNotIn(
            self.module_id,
            [mod.mod_id for mod in self.manager.subscriptions[MT_TEST_MESSAGE]],
            msg="Module id found in TEST_MESSAGE subscriptions.",
        )
        self.assertNotIn(
            self.module_id,
            [mod.mod_id for mod in self.manager.subscriptions[MT_TEST_MESSAGE2]],
            msg="Module id found in MT_TEST_MESSAGE2 subscriptions.",
        )

    def test_whenClientContextPauses_clientIsRemovedFromMessageType(self):
        """
        Test if client is unsubscribed from message type when pause is called.
        """
        # Arrange
        self.client.connect(server_name=self.addr)
        wait_for_message()
        with self.client.subscription_context([MT_TEST_MESSAGE]):
            wait_for_message()

            # Act
            with self.client.paused_subscription_context([MT_TEST_MESSAGE]):
                wait_for_message()

                # Assert
                self.assertNotIn(
                    self.module_id,
                    [mod.mod_id for mod in self.manager.subscriptions[MT_TEST_MESSAGE]],
                    msg="Module id found in TEST_MESSAGE subscriptions.",
                )

    def test_whenClientContextPausesMulti_clientIsRemovedFromMessageTypes(self):
        """
        Test if client is unsubscribed from message types when multiple message
            types are sent to pause.
        """
        # Arrange
        self.client.connect(server_name=self.addr)
        wait_for_message()
        with self.client.subscription_context([MT_TEST_MESSAGE, MT_TEST_MESSAGE2]):
            wait_for_message()

            # Act
            with self.client.paused_subscription_context(
                [MT_TEST_MESSAGE, MT_TEST_MESSAGE2]
            ):
                wait_for_message()

                # Assert
                self.assertNotIn(
                    self.module_id,
                    [mod.mod_id for mod in self.manager.subscriptions[MT_TEST_MESSAGE]],
                    msg="Module id found in TEST_MESSAGE subscriptions.",
                )
                self.assertNotIn(
                    self.module_id,
                    [
                        mod.mod_id
                        for mod in self.manager.subscriptions[MT_TEST_MESSAGE2]
                    ],
                    msg="Module id found in MT_TEST_MESSAGE2 subscriptions.",
                )

    def test_whenClientContextResumes_clientIsAddedToMessageType(self):
        """
        Test if client is subscribed to message type when resume is called.
        """
        # Arrange
        self.client.connect(server_name=self.addr)
        wait_for_message()
        with self.client.subscription_context([MT_TEST_MESSAGE]):
            wait_for_message()
            with self.client.paused_subscription_context([MT_TEST_MESSAGE]):
                wait_for_message()

            # Act
            # resumes subscription by exiting context above
            wait_for_message()

            # Assert
            self.assertIn(
                self.module_id,
                [mod.mod_id for mod in self.manager.subscriptions[MT_TEST_MESSAGE]],
                msg="Module id not found in TEST_MESSAGE subscriptions",
            )

    def test_whenClientContextResumesMulti_clientIsAddedToMessageType(self):
        """
        Test if client is subscribed to message types when multiple message
            types are sent to resume.
        """
        # Arrange
        self.client.connect(server_name=self.addr)
        wait_for_message()
        with self.client.subscription_context([MT_TEST_MESSAGE, MT_TEST_MESSAGE2]):
            wait_for_message()
            with self.client.paused_subscription_context(
                [MT_TEST_MESSAGE, MT_TEST_MESSAGE2]
            ):
                wait_for_message()

            # Act
            # resumes subscription by exiting context above
            wait_for_message()

            # Assert
            self.assertIn(
                self.module_id,
                [mod.mod_id for mod in self.manager.subscriptions[MT_TEST_MESSAGE]],
                msg="Module id not found in TEST_MESSAGE subscriptions",
            )
            self.assertIn(
                self.module_id,
                [mod.mod_id for mod in self.manager.subscriptions[MT_TEST_MESSAGE2]],
                msg="Module id not found in TEST_MESSAGE2 subscriptions",
            )


class TestSingleContextClient(unittest.TestCase):
    """
    Test interactions between a single client using client_context manager and message manager.
    """

    def setUp(self):
        self.port = random.randint(1000, 10000)  # random port
        self.addr = f"127.0.0.1:{self.port}"
        self.module_id = 11
        self.context_module_id = 12

        self.manager = MessageManager(
            ip_address="127.0.0.1",
            port=self.port,
            timecode=False,
            log_level=logging.ERROR,
            debug=False,
            send_msg_timing=True,
        )
        self.manager_thread = threading.Thread(
            target=self.manager.run,
        )
        self.manager_thread.start()
        wait_for_message()

    def tearDown(self):
        self.manager.close()
        self.manager_thread.join()

    def test_whenClientConnects_clientConnectsToMessageManager(self):
        """
        Test if client module is connected when connect is called.
        """
        # Arrange

        # Act
        with client_context(
            module_id=self.context_module_id, server_name=self.addr
        ) as client:
            wait_for_message()

            # Assert
            self.assertTrue(client.connected, msg="Client not connected.")
            self.assertIn(
                client.module_id,
                [mod.mod_id for mod in self.manager.modules.values()],
                msg="Module id not found in Messager Manager modules.",
            )

    def test_whenClientConnects_mod_id_is_correct(self):
        # Arrange

        # Act
        with client_context(
            module_id=self.context_module_id, server_name=self.addr
        ) as client:

            # Assert
            self.assertTrue(
                client.module_id == self.context_module_id,
                msg="Context client's module ID set incorrectly",
            )

    def test_whenClientContextSubscribes_clientIsSubscribedToMessageType(self):
        """
        Test if client is subscribed to message type when subscribe is called.
        """
        # Arrange

        # Act
        with client_context(
            module_id=self.context_module_id,
            server_name=self.addr,
            msg_list=[MT_TEST_MESSAGE],
        ) as client:
            wait_for_message()

            # Assert
            self.assertIn(
                client.module_id,
                [mod.mod_id for mod in self.manager.subscriptions[MT_TEST_MESSAGE]],
                msg="Module id not found in TEST_MESSAGE subscriptions.",
            )

    def test_whenClientContextSubscribesMulti_clientIsSubscribedToMessageTypes(
        self,
    ):
        """
        Test if client is subscribed to message types when multiple message
            types are sent to subscribe.
        """
        # Arrange

        # Act
        with client_context(
            module_id=self.context_module_id,
            server_name=self.addr,
            msg_list=[MT_TEST_MESSAGE, MT_TEST_MESSAGE2],
        ) as client:
            wait_for_message()

            # Assert
            self.assertIn(
                client.module_id,
                [mod.mod_id for mod in self.manager.subscriptions[MT_TEST_MESSAGE]],
                msg="Module id not found in TEST_MESSAGE subscriptions",
            )
            self.assertIn(
                client.module_id,
                [mod.mod_id for mod in self.manager.subscriptions[MT_TEST_MESSAGE2]],
                msg="Module id not found in TEST_MESSAGE2 subscriptions",
            )

    def test_whenClientContextUnsubscribes_clientIsUnsubscribedFromMessageType(
        self,
    ):
        """
        Test if client is unsubscribed from message type when unsubscribe is called.
        """
        # Arrange
        with client_context(
            module_id=self.context_module_id, server_name=self.addr
        ) as client:
            wait_for_message()
            with client.subscription_context([MT_TEST_MESSAGE]):
                wait_for_message()

            # Act
            # unsubscribes by exiting context above
            wait_for_message()

            # Assert
            self.assertNotIn(
                client.module_id,
                [mod.mod_id for mod in self.manager.subscriptions[MT_TEST_MESSAGE]],
                msg="Module id found in TEST_MESSAGE subscriptions.",
            )

    def test_whenClientContextUnsubscribesMulti_clientIsUnsubscribedFromMessageTypes(
        self,
    ):
        """
        Test if client is unsubscribed from message types when multiple message
            types are sent to unsubscribe.
        """
        # Arrange
        with client_context(
            module_id=self.context_module_id, server_name=self.addr
        ) as client:
            wait_for_message()
            with client.subscription_context([MT_TEST_MESSAGE, MT_TEST_MESSAGE2]):
                wait_for_message()

            # Act
            # unsubscribes by exiting context above
            wait_for_message()

            # Assert
            self.assertNotIn(
                client.module_id,
                [mod.mod_id for mod in self.manager.subscriptions[MT_TEST_MESSAGE]],
                msg="Module id found in TEST_MESSAGE subscriptions.",
            )
            self.assertNotIn(
                client.module_id,
                [mod.mod_id for mod in self.manager.subscriptions[MT_TEST_MESSAGE2]],
                msg="Module id found in MT_TEST_MESSAGE2 subscriptions.",
            )

    def test_whenClientContextPauses_clientIsRemovedFromMessageType(self):
        """
        Test if client is unsubscribed from message type when pause is called.
        """
        # Arrange
        with client_context(
            module_id=self.context_module_id,
            server_name=self.addr,
            msg_list=[MT_TEST_MESSAGE],
        ) as client:
            wait_for_message()

            # Act
            with client.paused_subscription_context([MT_TEST_MESSAGE]):
                wait_for_message()

                # Assert
                self.assertNotIn(
                    client.module_id,
                    [mod.mod_id for mod in self.manager.subscriptions[MT_TEST_MESSAGE]],
                    msg="Module id found in TEST_MESSAGE subscriptions.",
                )

    def test_whenClientContextPausesMulti_clientIsRemovedFromMessageType(self):
        """
        Test if client is unsubscribed from message types when multiple message
            types are sent to pause.
        """
        # Arrange
        with client_context(
            module_id=self.context_module_id,
            server_name=self.addr,
            msg_list=[MT_TEST_MESSAGE, MT_TEST_MESSAGE2],
        ) as client:
            wait_for_message()

            # Act
            with client.paused_subscription_context(
                [MT_TEST_MESSAGE, MT_TEST_MESSAGE2]
            ):
                wait_for_message()

                # Assert
                self.assertNotIn(
                    client.module_id,
                    [mod.mod_id for mod in self.manager.subscriptions[MT_TEST_MESSAGE]],
                    msg="Module id found in TEST_MESSAGE subscriptions.",
                )
                self.assertNotIn(
                    client.module_id,
                    [
                        mod.mod_id
                        for mod in self.manager.subscriptions[MT_TEST_MESSAGE2]
                    ],
                    msg="Module id found in TEST_MESSAGE2 subscriptions.",
                )

    def test_whenClientContextResumes_clientIsAddedToMessageType(self):
        """
        Test if client is subscribed to message type when resume is called.
        """
        # Arrange
        with client_context(
            module_id=self.context_module_id,
            server_name=self.addr,
            msg_list=[MT_TEST_MESSAGE],
        ) as client:
            wait_for_message()
            with client.paused_subscription_context([MT_TEST_MESSAGE]):
                wait_for_message()

            # Act
            # resumes subscription by exiting context above
            wait_for_message()

            # Assert
            self.assertIn(
                client.module_id,
                [mod.mod_id for mod in self.manager.subscriptions[MT_TEST_MESSAGE]],
                msg="Module id not found in TEST_MESSAGE subscriptions",
            )

    def test_whenClientContextResumesMulti_clientIsAddedToMessageType(self):
        """
        Test if client is subscribed to message types when multiple message
            types are sent to resume.
        """
        # Arrange
        with client_context(
            module_id=self.context_module_id,
            server_name=self.addr,
            msg_list=[MT_TEST_MESSAGE, MT_TEST_MESSAGE2],
        ) as client:
            wait_for_message()
            with client.paused_subscription_context(
                [MT_TEST_MESSAGE, MT_TEST_MESSAGE2]
            ):
                wait_for_message()

            # Act
            # resumes subscription by exiting context above
            wait_for_message()

            # Assert
            self.assertIn(
                client.module_id,
                [mod.mod_id for mod in self.manager.subscriptions[MT_TEST_MESSAGE]],
                msg="Module id not found in TEST_MESSAGE subscriptions",
            )
            self.assertIn(
                client.module_id,
                [mod.mod_id for mod in self.manager.subscriptions[MT_TEST_MESSAGE2]],
                msg="Module id not found in TEST_MESSAGE2 subscriptions",
            )



from .data_logger import DataLogger


def _load_default_formatters():
    from .formatters.quicklogger import QLFormatter
    from .formatters.raw import RawFormatter
    from .formatters.json import JsonFormatter
    from .formatters.msg_headers import MsgHeaderFormatter
    from .data_formatter import add_formatter

    # Add Builtin Formatters
    add_formatter(QLFormatter)
    add_formatter(RawFormatter)
    add_formatter(JsonFormatter)
    add_formatter(MsgHeaderFormatter)


_load_default_formatters()



import pathlib
import logging
import threading
import time

from ..message import Message
from .metadata import LoggingMetadata
from .data_set import DataSet
from .exceptions import (
    DataCollectionThreadError,
    DataCollectionFullError,
    BasePathNotFound,
)

from typing import Optional


class DataCollection:
    MAX_DATA_SETS = 6
    WRITE_PERIOD = 15.0

    def __init__(
        self,
        name: str,
        base_path: str,
        dir_fmt: str,
        metadata: LoggingMetadata,
        use_thread: bool = True,
    ):
        self._dead = False

        self.logger = logging.getLogger("data_logger").getChild(name)
        self.logger.name = name
        self._recording = False
        self._paused = False
        self._close = False

        self._elapsed_time = 0.0
        self.ref_time = -1
        self.start_time = -1
        self.next_write = -1.0

        self.name = name

        self.dir_fmt = dir_fmt

        self.metadata = metadata
        self.datasets: list[DataSet] = []
        self.write_to_disk = threading.Event()
        self.write_finished = threading.Event()

        ex_base_path = self.metadata.expand_format(base_path)
        p = pathlib.Path(ex_base_path)
        if not p.exists() or not p.is_dir():
            raise BasePathNotFound(
                f"The base_path directory specified does not exist: {base_path}"
            )

        self.base_path = p
        self.save_path = p

        if use_thread:
            self.use_thread = use_thread
            self.write_thread = threading.Thread(target=self.write)
            self.write_thread.start()

    def __del__(self):
        if self._dead:
            return
        self.close()

    def close(self):
        self._close = True
        if self.use_thread:
            self.logger.info(f"Waiting for {self.name} write thread to close...")
            self.write_thread.join()

        for ds in self.datasets:
            ds.close()

        self._dead = True

    def pause(self):
        elapsed = self.elapsed_time
        self._elapsed_time = elapsed
        self._paused = True

    def resume(self):
        self._paused = False
        self.ref_time = time.time()

    @property
    def elapsed_time(self) -> float:
        if self.stopped:
            return 0

        if self.paused:
            return self._elapsed_time

        return self._elapsed_time + (time.time() - self.ref_time)

    @property
    def total_elapsed_time(self) -> float:
        return time.time() - self.start_time

    @property
    def paused(self) -> bool:
        return self._paused

    @property
    def stopped(self) -> bool:
        return not self._recording

    def start(self):
        if self.use_thread and not self.write_thread.is_alive():
            raise DataCollectionThreadError(
                "Data collection write thread is no longer active. Reset the logger."
            )

        # Expand collection save directory
        save_dir = self.metadata.expand_format(self.dir_fmt)

        p = self.base_path
        self.save_path = p.joinpath(save_dir)

        self.save_path.mkdir(parents=True, exist_ok=True)
        self.logger.info(f"Saving collection: {self.name} -> {self.save_path}")

        self.start_time = time.time()
        self.ref_time = self.start_time

        # Start up all the data set files
        for ds in self.datasets:
            ds.start(self.save_path)

        self.next_write = DataCollection.WRITE_PERIOD
        self._recording = True
        self._paused = False

    def stop(self):
        self.logger.info(f"Stopping collection: {self.name} -> {self.save_path}")
        # Check if we are currently writing some data
        if self.write_to_disk.is_set():
            while not self.write_finished.wait(0.250):
                pass

        self.write_to_disk.clear()
        self.write_finished.clear()

        # Close all the Data Set files in the collection
        for ds in self.datasets:
            ds.collection_stopped = True
            ds.stop()
            ds.close()

        self.start_time = -1
        self.ref_time = -1
        self._recording = False
        self._paused = False
        self.next_write = -1.0

        self.logger.info(f"Stopped collection: {self.name}")

    def update_metadata(self, metadata: LoggingMetadata):
        self.metadata = metadata

        for ds in self.datasets:
            ds.metadata = metadata

    def update(self, msg: Optional[Message]):
        if self._paused or not self._recording:
            return

        if self.use_thread and not self.write_thread.is_alive():
            raise DataCollectionThreadError(
                "Data collection write thread is no longer active. Reset the logger."
            )

        write = False
        elapsed = self.elapsed_time
        for ds in self.datasets:
            if msg:
                if ds.all_sub or msg.type_id in ds.msg_types:
                    ds.rbuf.append(msg)

            if elapsed > ds.next_subdivide:
                ds.next_subdivide = elapsed + ds.subdivide_interval
                ds.subdivide_flag = True
                write = True

        if elapsed > self.next_write:
            write = True

        if write:
            if self.write_to_disk.is_set():
                self.logger.warning("Unable to write fast enough.")
            else:
                self.next_write = elapsed + DataCollection.WRITE_PERIOD
                self.trigger_write()

    def trigger_write(self):
        for ds in self.datasets:
            ds.stage_for_write()

        self.write_finished.clear()
        self.write_to_disk.set()

        if not self.use_thread:
            self.blocking_write()
        self.logger.debug("Writing data set buffers to disk.")

    def add_data_set(self, data_set: DataSet):
        if data_set.name in [ds.name for ds in self.datasets]:
            # Remove the entry that we are overwriting
            self.datasets = [ds for ds in self.datasets if ds.name != data_set.name]
            self.logger.info(f"Updated data set: '{data_set.name}'")
        elif len(self.datasets) == DataCollection.MAX_DATA_SETS:
            raise DataCollectionFullError(
                "Collection has maximum allowed data sets configured."
            )
        else:
            self.logger.info(f"Added data set: {self.name}->{data_set.name}")

        self.datasets.append(data_set)

    def rm_data_set(self, name: str):
        self.datasets = [ds for ds in self.datasets if ds.name != name]
        self.logger.info(f"Removed data set: '{name}'")

    def write(self):
        """Write Threads main function"""
        try:
            while not self._close:
                if self.write_to_disk.wait(0.5):
                    for ds in self.datasets:
                        ds.write()
                    self.write_to_disk.clear()
                    self.write_finished.set()
        except KeyboardInterrupt:
            pass
        finally:
            print("Collection write thread exited.")

    def blocking_write(self):
        """Blocking write without bg thread"""
        if self.write_to_disk.wait(0.5):
            for ds in self.datasets:
                ds.write()
            self.write_to_disk.clear()
            self.write_finished.set()



import pyrtma

from ..data_formatter import DataFormatter
from typing import ClassVar


class JsonFormatter(DataFormatter):
    name: ClassVar[str] = "json"
    mode: ClassVar[str] = "wt"
    ext: ClassVar[str] = ".json"

    def format_message(self, msg: pyrtma.Message) -> str:
        return msg.to_json(minify=True) + "\n"



import pathlib

from typing import Union
from pyrtma.parser import (
    Parser,
    ConstantExpr,
    ConstantString,
    TypeAlias,
    MT,
    MID,
    HID,
    MDF,
    SDF,
)
from pyrtma.__version__ import __version__

# Native C types that are supported
type_map = {
    "char": '""',
    "string": '""',  # not really a C type, used to distinguish char from string (length > 1) in js
    "unsigned char": 0,
    "byte": 0,
    "int": 0,
    "signed int": 0,
    "unsigned int": 0,
    "unsigned": 0,
    "short": 0,
    "signed short": 0,
    "unsigned short": 0,
    "long": 0,
    "signed long": 0,
    "unsigned long": 0,
    "long long": 0,
    "signed long long": 0,
    "unsigned long long": 0,
    "float": 0,
    "double": 0,
    "uint8": 0,
    "uint16": 0,
    "uint32": 0,
    "uint64": 0,
    "int8": 0,
    "int16": 0,
    "int32": 0,
    "int64": 0,
}

TAB = "  "  # two spaces is preferred indent in js


def pad(indent: int) -> str:
    return indent * TAB


class JSDefCompiler:
    def __init__(self, parser: Parser, debug: bool = False):
        self.debug = debug
        self.parser = parser

    def generate_constant(self, c: ConstantExpr):
        return f"RTMA.constants.{c.name} = {c.value};\n"

    def generate_string_constant(self, c: ConstantString):
        return f"RTMA.constants.{c.name} = {c.value};\n"

    def generate_prop(self, name: str, value: str):
        return f"{name}: {value}"

    def generate_msg_type_id(self, mt: MT) -> str:
        return f"RTMA.MT.{mt.name} = {mt.value};\n"

    def generate_host_id(self, hid: HID) -> str:
        return f"RTMA.HID.{hid.name} = {hid.value};\n"

    def generate_module_id(self, mid: MID) -> str:
        return f"RTMA.MID.{mid.name} = {mid.value};\n"

    def generate_hash_id(self, mdf: Union[MDF, SDF]) -> str:
        return f'RTMA.HASH.{mdf.name} = "{mdf.hash[:8]}";\n'

    def generate_type_alias(self, td: TypeAlias) -> str:
        if td.type_name in type_map.keys():
            clean_name = td.type_name.replace(" ", "_")
            return f"RTMA.aliases.{td.name} = type_map.{clean_name}();\n"

        if td.type_name in self.parser.aliases.keys():
            return f"RTMA.aliases.{td.name} = RTMA.aliases.{td.type_name};\n"

        if td.type_name in self.parser.struct_defs.keys():
            return f"RTMA.SDF.{td.name} = RTMA.SDF.{td.type_name};\n"

        if td.type_name in self.parser.message_defs.keys():
            return f"RTMA.MDF.{td.name} = RTMA.MDF.{td.type_name};\n"

        raise RuntimeError(f"No type found for alias: {td.name}")

    def generate_obj(self, struct: Union[SDF, MDF]) -> str:
        if isinstance(struct, MDF):
            top_field = "MDF"
        elif isinstance(struct, SDF):
            top_field = "SDF"

        tabs = TAB * 2
        num_fields = len(struct.fields)

        if num_fields == 0:
            return f"RTMA.{top_field}.{struct.name} = () => {{ return {{}} }};"

        s = f"RTMA.{top_field}.{struct.name} = () => {{\n{TAB}return {{\n"

        for n, field in enumerate(struct.fields, start=1):
            s += tabs
            if field.type_name in type_map.keys():
                clean_name = field.type_name.replace(" ", "_")
                ftype = f"type_map.{clean_name}"
            elif field.type_name in self.parser.message_defs.keys():
                ftype = f"RTMA.MDF.{field.type_name}"
            elif field.type_name in self.parser.struct_defs.keys():
                ftype = f"RTMA.SDF.{field.type_name}"
            elif field.type_name in self.parser.aliases.keys():
                ftype = f"RTMA.aliases.{field.type_name}"
            else:
                raise RuntimeError(f"Unknown field name {field.name} in {struct.name}")

            comment = ""  # reserved for possible future use

            if field.length is not None:
                if field.type_name == "char" and field.length > 1:
                    ftype = "type_map.string"
                    s += f"{field.name}: {ftype}({field.length})"
                else:
                    s += f"{field.name}: Array({field.length}).fill({ftype}())"
            else:
                s += f"{field.name}: {ftype}()"

            if n < num_fields:
                s += f",{comment}\n"
            else:
                s += f"{comment}\n"

        s += f"{TAB}}}\n"
        s += "};"

        return s

    def generate(self, out_filepath: pathlib.Path):
        with open(out_filepath, mode="w") as f:
            # Version comment
            f.write(
                f"// This message def file was auto-generated by pyrtma.compile version {__version__}\n\n"
            )

            # Exports
            f.write("export { RTMA } ;\n\n")

            # Generate Type Map Helper
            f.write("// Type Map Default Values\n")
            f.write("const type_map = {};\n")
            for name, value in type_map.items():
                name = name.replace(" ", "_")
                args = "length" if name == "string" else ""
                f.write(f"type_map.{name} = ({args}) => {value};\n")
            f.write("\n")

            # Top-Level RTMA object
            f.write("// Top-Level RTMA object\n")
            f.write("const RTMA = {};\n")
            f.write(f'RTMA.COMPILED_PYRTMA_VERSION = "{__version__}";\n\n')

            # RTMA.constants
            f.write("// Constants\n")
            f.write("RTMA.constants =  {};\n")
            obj: Union[ConstantExpr, ConstantString, HID, MID, TypeAlias, SDF, MT, MDF]
            for obj in self.parser.constants.values():
                f.write(self.generate_constant(obj))
            f.write("\n")

            f.write("// String Constants\n")
            for obj in self.parser.string_constants.values():
                f.write(self.generate_string_constant(obj))
            f.write("\n")

            # RTMA.typedefs
            f.write("// Type Aliases\n")
            f.write("RTMA.aliases =  {};\n\n")
            for obj in self.parser.aliases.values():
                f.write(self.generate_type_alias(obj))
            f.write("\n")

            # RTMA.HID
            f.write("// Host IDs\n")
            f.write("RTMA.HID =  {};\n\n")
            for obj in self.parser.host_ids.values():
                f.write(self.generate_host_id(obj))
            f.write("\n")

            # RTMA.MID
            f.write("// Module IDs\n")
            f.write("RTMA.MID =  {};\n\n")
            for obj in self.parser.module_ids.values():
                f.write(self.generate_module_id(obj))
            f.write("\n")

            # RTMA.MT
            f.write("// Message Type IDs\n")
            f.write("RTMA.MT = {};\n\n")
            for obj in self.parser.message_ids.values():
                f.write(self.generate_msg_type_id(obj))
            f.write("\n")

            # RTMA.SDF
            f.write("// Struct Definitions\n")
            f.write("RTMA.SDF = {};\n\n")
            for obj in self.parser.struct_defs.values():
                f.write(self.generate_obj(obj))
                f.write("\n\n")

            # RTMA.MDF
            f.write("// Message Definitions\n")
            f.write("RTMA.MDF = {};\n\n")
            for obj in self.parser.message_defs.values():
                f.write(self.generate_obj(obj))
                f.write("\n\n")

            # RTMA.MDF
            f.write("// Message Definition Hashes\n")
            f.write("RTMA.HASH = {};\n\n")
            for obj in self.parser.message_defs.values():
                f.write(self.generate_hash_id(obj))
